/********************************************************************************
*                                                                               *
*                             B i t m a p    O b j e c t                        *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXBitmap.cpp,v 1.14 1999/01/07 03:38:42 jeroen Exp $                     *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDrawable.h"
#include "FXBitmap.h"



/*
  Note:
  - Make this work.
  - Try eliminate temp copy:- slap pixels into XImage directly, if possible...
  - Perhaps enforce system-native padding necessary for the above.
  - Our bitmap data is 01234567, i.e. LS-BIT first; byte order ditto.
  - Issue: should FXBitmap return the DC for drawing onto the X-Server resident
    pixmap, or the client-side image bits?
    
    My idea is it should be the latter:
    
      - Allows even richer set of drawing primitives, as everything is
        drawn in software.
      - Very useful to generate off-screen renderings, e.g. during printing.
      - Allows for building and running true-color drawing programs on
        low-end graphics hardware.
      - The only drawback I can see is it will be a fairly large implementation
        effort...
      
*/

/*******************************************************************************/

// Bit-reversal
static const FXuchar reverse_byte[256]={
  0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
  0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
  0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
  0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
  0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
  0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
  0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
  0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
  0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
  0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
  0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
  0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
  0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
  0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
  0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
  0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
  0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
  0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
  0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
  0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
  0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
  0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
  0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
  0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
  0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
  0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
  0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
  0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
  0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
  0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
  0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
  0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
  };


/*******************************************************************************/

// Object implementation
FXIMPLEMENT(FXBitmap,FXDrawable,NULL,0)


// For deserialization
FXBitmap::FXBitmap(){
  data=NULL;
  options=0;
  }


// Initialize
FXBitmap::FXBitmap(FXApp* a,const void *pix,FXuint opts,FXint w,FXint h):FXDrawable(a,w,h,1){
  FXTRACE((100,"FXBitmap::FXBitmap %08x\n",this));
  if(pix) opts&=~BITMAP_OWNED;
  data=(FXuchar*)pix;
  options=opts;
  }


// Create bitmap
void FXBitmap::create(){
  if(!xid){
    FXTRACE((100,"%s::create %08x\n",getClassName(),this));
    
#ifndef FX_NATIVE_WIN32
    
    // App should exist
    if(!getDisplay()){ fxerror("%s::create: trying to create image before opening display.\n",getClassName()); }
  
    // Make pixmap
    xid=XCreatePixmap(getDisplay(),XDefaultRootWindow(getDisplay()),width,height,1);
    if(!xid){ fxerror("%s::create: unable to create image.\n",getClassName()); }
  
#else

    // Make an empty bitmap
    xid=CreateBitmap(width,height,1,1,NULL);
    if(!xid){ fxerror("%s::create: unable to create image.\n",getClassName()); }

#endif

    // Render pixels
    render();

    // Zap data
    if(!(options&BITMAP_KEEP) && (options&BITMAP_OWNED)){
      options&=~BITMAP_OWNED;
      FXFREE(&data);
      }
    }
  }


// Destroy image
void FXBitmap::destroy(){
  if(xid){
    FXTRACE((100,"%s::destroy %08x\n",getClassName(),this));
#ifndef FX_NATIVE_WIN32
    XFreePixmap(getDisplay(),xid);
#else
    DeleteObject(xid);
#endif
    xid=0;
    }
  }


// Return DC for this surface
FXDC* FXBitmap::DC() const {
  return getApp()->bitmapdc;
  }

#ifndef FX_NATIVE_WIN32

// Render bits
void FXBitmap::render_bits(XImage *xim,FXuchar *img){
  register int size=xim->bytes_per_line*height;
  register FXuchar *pix=(FXuchar*)xim->data;
  register int i;
  
  // Most significant bit first
  if(xim->bitmap_bit_order==MSBFirst){
    for(i=0; i<size; i++) pix[i]=reverse_byte[img[i]];
    }
  
  // Least significant bit first
  else{
    memcpy(pix,img,size);
    }
  }

#endif

// Render into pixmap
void FXBitmap::render(){
  
  FXTRACE((100,"%s::render bitmap %0x8\n",getClassName(),this));
  
  // Do it for X-Windows
#ifndef FX_NATIVE_WIN32
  
 // XImage xim;
  register XImage *xim=NULL;
  XGCValues values;
  GC gc;

  // Can not render before creation
  if(!xid){ fxerror("%s::render: trying to render bitmap before it has been created.\n",getClassName()); }

  // Check for legal size
  if(width<2 || height<2){ fxerror("%s::render: illegal bitmap size.\n",getClassName()); }

  // Just leave if black if no data
  if(data){
    
    // Make GC
    values.foreground=0xffffffff;
    values.background=0;
    gc=XCreateGC(getDisplay(),xid,GCForeground|GCBackground,&values);
    
    xim=XCreateImage(getDisplay(),DefaultVisual(getDisplay(),DefaultScreen(getDisplay())),1,XYBitmap,0,NULL,width,height,8,(width+7)>>3);
    if(!xim){ fxerror("%s::render: unable to render image.\n",getClassName()); }

    // Try create temp pixel store
    xim->data=(char*)malloc(xim->bytes_per_line*height);

    // Failed completely
    if(!xim->data){ fxerror("%s::render: unable to allocate memory.\n",getClassName()); }
    
    // Render bits into server-formatted bitmap
    render_bits(xim,data);
    
    // Blast the image
    XPutImage(getDisplay(),xid,gc,xim,0,0,0,0,width,height);
#ifndef WIN32
    //// Need to use something other than malloc for WIN32....
    XDestroyImage(xim);
#endif
    
    // We're done
    XFreeGC(getDisplay(),gc);
    }
  
#else
  
  // Can not render before creation
  if(!xid){ fxerror("%s::render: trying to render bitmap before it has been created.\n",getClassName()); }

  // Check for legal size
  if(width<2 || height<2){ fxerror("%s::render: illegal bitmap size.\n",getClassName()); }

  // Just leave if black if no data
  if(data){
    BITMAP bm;
    GetObject(xid,sizeof(BITMAP),&bm);
    DWORD cBytes=bm.bmHeight*bm.bmWidthBytes;
    SetBitmapBits((HBITMAP)xid,cBytes,data);
    }

#endif
  }


// Save pixel data only
void FXBitmap::savePixels(FXStream& store) const {
  FXuint size=height*((width+7)>>3);
  store.save(data,size);
  }


// Load pixel data only
void FXBitmap::loadPixels(FXStream& store){
  FXuint size=height*((width+7)>>3);
  if(options&BITMAP_OWNED) FXFREE(&data);
  FXMALLOC(&data,FXuchar,size);
  store.load(data,size);
  options|=BITMAP_OWNED;
  }


// Save data
void FXBitmap::save(FXStream& store) const {
  FXuchar haspixels=(data!=NULL);
  FXDrawable::save(store);
  store << options;
  store << haspixels;
  if(haspixels) savePixels(store);
  }


// Load data
void FXBitmap::load(FXStream& store){
  FXuchar haspixels;
  FXDrawable::load(store);
  store >> options;
  store >> haspixels;
  if(haspixels) loadPixels(store);
  }


// Clean up
FXBitmap::~FXBitmap(){
  FXTRACE((100,"FXBitmap::~FXBitmap %08x\n",this));
  if(xid){
#ifndef FX_NATIVE_WIN32
    XFreePixmap(getDisplay(),xid);
#else
    DeleteObject(xid);
#endif
    }
  if(options&BITMAP_OWNED){FXFREE(&data);}
  data=(FXuchar*)-1;
  }
  
