/********************************************************************************
*                                                                               *
*                         C u r s o r - O b j e c t                             *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXCursor.cpp,v 1.18 1999/01/13 11:13:51 jlj Exp $                     *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXCursor.h"

/*
  To do:
  - Add subclass to read in .CUR file (similar to FXGIFIcon reading
    e.g. a GIF icon) from an FXStream.
  - Cursors should be 32x32, as that's the only option under Windows!
*/


/*******************************************************************************/

// Object implementation
FXIMPLEMENT(FXCursor,FXId,NULL,0)


// Deserialization
FXCursor::FXCursor(){
  source=NULL;
  mask=NULL;
  width=0;
  height=0;
  hotx=-1;
  hoty=-1;
  glyph=0;
  owned=FALSE;
  }


// Make stock cursor
FXCursor::FXCursor(FXApp* a,FXStockCursor curid):FXId(a,0){
  FXTRACE((100,"FXCursor::FXCursor %08x\n",this));
  source=NULL;
  mask=NULL;
  width=0;
  height=0;
  hotx=-1;
  hoty=-1;
  glyph=(FXuchar)curid;   // Got to do this, we can't serialize enums (don't know how big they are!)
  owned=FALSE;
  }


// Make cursor from source and mask
FXCursor::FXCursor(FXApp* a,const void* src,const void* msk,FXint w,FXint h,FXint hx,FXint hy):FXId(a,0){
  FXTRACE((100,"FXCursor::FXCursor %08x\n",this));
  source=(FXuchar*)src;
  mask=(FXuchar*)msk;
  width=w;
  height=h;
  hotx=hx;
  hoty=hy;
  glyph=0;
  owned=FALSE;
  }
  

// Create cursor
void FXCursor::create(){
  if(!xid){
    FXTRACE((100,"%s::create %08x\n",getClassName(),this));
    
    // X-Windows
#ifndef FX_NATIVE_WIN32
    const FXuint stock[]={XC_top_left_arrow,XC_top_left_arrow,XC_arrow,XC_xterm,XC_watch};
    Pixmap srcpix,mskpix;
    XColor color[2];
    
    // App should exist
    if(!getDisplay()){ fxerror("%s::create: trying to create cursor before opening display.\n",getClassName()); }
  
    // Building custom cursor
    if(glyph==0){
      
      // Should have both source and mask
      if(!source || !mask){ fxerror("%s::create: cursor needs both source and mask.\n",getClassName()); }

      color[0].pixel=BlackPixel(getDisplay(),DefaultScreen(getDisplay()));
      color[1].pixel=WhitePixel(getDisplay(),DefaultScreen(getDisplay()));
      color[0].flags=DoRed|DoGreen|DoBlue;
      color[1].flags=DoRed|DoGreen|DoBlue;
      XQueryColors(getDisplay(),DefaultColormap(getDisplay(),DefaultScreen(getDisplay())),color,2);

      // Create pixmaps for source and mask
      srcpix=XCreateBitmapFromData(getDisplay(),XDefaultRootWindow(getDisplay()),(char*)source,width,height);
      mskpix=XCreateBitmapFromData(getDisplay(),XDefaultRootWindow(getDisplay()),(char*)mask,width,height);

      // Create cursor
      xid=XCreatePixmapCursor(getDisplay(),srcpix,mskpix,&color[0],&color[1],hotx,hoty);
      if(!xid){ fxerror("%s::create: unable to create cursor.\n",getClassName()); }

      // No longer needed
      XFreePixmap(getDisplay(),srcpix);
      XFreePixmap(getDisplay(),mskpix);
      }
    
    // Building stock cursor
    else{
      FXASSERT(glyph<ARRAYNUMBER(stock));
      xid=XCreateFontCursor(getDisplay(),stock[glyph]);
      }
    
    // MS-Windows
#else
    const LPCTSTR stock[]={IDC_ARROW,IDC_ARROW,IDC_ARROW,IDC_IBEAM,IDC_IBEAM};

    // Building custom cursor
    if(glyph==0){
      // Should have both source and mask
      if(!source || !mask){ fxerror("%s::create: cursor needs both source and mask.\n",getClassName()); }

      // Let's hope it's the correct size!
      FXASSERT(GetSystemMetrics(SM_CXCURSOR)==width);
      FXASSERT(GetSystemMetrics(SM_CYCURSOR)==height);

      // Windows cursor bitmaps are reversed left-to-right
      FXASSERT(GetSystemMetrics(SM_CXCURSOR)==32); // Following loops assume 4 bytes per row
      int offset=0;
      BYTE tmpsrc[4],tmpmsk[4];
      for(int h=0; h<height; h++){
	for(int i=0; i<4; i++){
	  tmpsrc[i]=source[offset+i];
	  tmpmsk[i]=mask[offset+i];
	  }
	for(int j=0; j<4; j++){
	  source[offset+j]=tmpsrc[3-j];
	  mask[offset+j]=tmpmsk[3-j];
	  }
	offset+=4;
	}

      // Need to alter bitmasks a bit
      HDC hdcDesktop=GetDC(GetDesktopWindow());
      
      HDC hdcSrc=CreateCompatibleDC(hdcDesktop);
      HBITMAP hSrcBitmap=CreateBitmap(width,height,1,1,source);
      SelectObject(hdcSrc,hSrcBitmap);

      HDC hdcMask=CreateCompatibleDC(hdcDesktop);
      HBITMAP hMaskBitmap=CreateBitmap(width,height,1,1,mask);
      SelectObject(hdcMask,hMaskBitmap);

      BitBlt(hdcSrc,0,0,width,height,hdcMask,0,0,SRCERASE);
      BitBlt(hdcMask,0,0,width,height,hdcMask,0,0,DSTINVERT);

      DeleteDC(hdcSrc);
      DeleteDC(hdcMask);

      // Grab the modified bitmask arrays
      BITMAPINFO bmi;
      bmi.bmiHeader.biSize=sizeof(BITMAPINFO);
      bmi.bmiHeader.biWidth=width;
      bmi.bmiHeader.biHeight=height;
      bmi.bmiHeader.biPlanes=1;
      bmi.bmiHeader.biBitCount=1;
      bmi.bmiHeader.biCompression=0;
      GetDIBits(hdcDesktop,hMaskBitmap,0,height,mask,&bmi,DIB_RGB_COLORS);
      GetDIBits(hdcDesktop,hSrcBitmap,0,height,source,&bmi,DIB_RGB_COLORS);

      // Done with these...
      ReleaseDC(GetDesktopWindow(),hdcDesktop);
      DeleteObject(hMaskBitmap);
      DeleteObject(hSrcBitmap);

      if(GetSystemMetrics(SM_CXCURSOR)==width && GetSystemMetrics(SM_CYCURSOR)==height)
        xid=CreateCursor(getApp()->hInstance,width-hotx,height-hoty,width,height,mask,source);
      else
        xid=LoadCursor(NULL,IDC_ARROW);
      }
    
    // Building stock cursor
    else{
      FXASSERT(glyph<ARRAYNUMBER(stock));
      xid=LoadCursor(NULL,stock[glyph]);
      }
#endif
    }
  }


// Destroy cursor
void FXCursor::destroy(){
  if(xid){
    FXTRACE((100,"%s::destroy %08x\n",getClassName(),this));
#ifndef FX_NATIVE_WIN32
    XFreeCursor(getApp()->display,xid);
#else
    DestroyCursor((HCURSOR)xid);
#endif
    xid=0;
    }
  }


// Save pixel data only
void FXCursor::savePixels(FXStream& store) const {
  store << width << height;
  store << hotx << hoty;
  store << glyph;
  if(glyph==0){
    FXuint size=height*((width+7)>>3);
    store.save(source,size);
    store.save(mask,size);
    }
  }


// Load pixel data only
void FXCursor::loadPixels(FXStream& store){
  store >> width >> height;
  store >> hotx >> hoty;
  store >> glyph;
  if(glyph==0){
    if(owned){FXFREE(&source);FXFREE(&mask);}
    FXuint size=height*((width+7)>>3);
    FXMALLOC(&source,FXuchar,size);
    FXMALLOC(&mask,FXuchar,size);
    store.load(source,size);
    store.load(mask,size);
    owned=TRUE;
    }
  }


// Save object to stream
void FXCursor::save(FXStream& store) const {
  FXId::save(store);
  savePixels(store);
  }


// Load object from stream
void FXCursor::load(FXStream& store){
  FXId::load(store);
  loadPixels(store);
  }  


// Clean up
FXCursor::~FXCursor(){
  FXTRACE((100,"FXCursor::~FXCursor %08x\n",this));
#ifndef FX_NATIVE_WIN32
  if(xid){XFreeCursor(getApp()->display,xid);}
#else
  if(xid){DestroyCursor((HCURSOR)xid);}
#endif
  if(owned){FXFREE(&source);FXFREE(&mask);}
  source=(FXuchar*)-1;
  mask=(FXuchar*)-1;
  }
  

