/********************************************************************************
*                                                                               *
*                             M e n u   O b j e c t s                           *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXMenu.cpp,v 1.73 1999/01/14 06:07:47 jeroen Exp $                       *
********************************************************************************/
#include "fxdefs.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXShell.h"
#include "FXPopup.h"
#include "FXMenu.h"

/*
  Notes:
  - Mnemonics, accelerators.
  - Help text from constructor is third part; second part should be
    accelerator key combination.
  - When menu label changes, hotkey might have to be adjusted.
  - Fix it so menu stays up when after Alt-F, you press Alt-E.
  - MenuItems should be derived from FXLabel.
  - FXMenuCascade should send ID_POST/IDUNPOST to self.
  - Look into SEL_FOCUS_SELF some more...
*/


#define LEADSPACE   22
#define TRAILSPACE  16

/*******************************************************************************/

// Map
FXDEFMAP(FXMenuSeparator) FXMenuSeparatorMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXMenuSeparator::onPaint),
  };


// Object implementation
FXIMPLEMENT(FXMenuSeparator,FXFrame,FXMenuSeparatorMap,ARRAYNUMBER(FXMenuSeparatorMap))


// Separator item
FXMenuSeparator::FXMenuSeparator(FXComposite* p,FXuint opts):
  FXFrame(p,opts,0,0,0,0){ 
  defaultCursor=getApp()->rarrowCursor;
  }


// Handle repaint 
long FXMenuSeparator::onPaint(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  dc->begin(this);
  dc->setForeground(backColor);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  dc->fillRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  dc->setForeground(shadowColor);
  dc->drawLine(1,0,width-1,0);
  dc->setForeground(hiliteColor);
  dc->drawLine(1,1,width-1,1);
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Get default size
FXint FXMenuSeparator::getDefaultWidth(){ 
  return LEADSPACE+TRAILSPACE;
  }


FXint FXMenuSeparator::getDefaultHeight(){ 
  return 2; 
  }


/*******************************************************************************/

// Map
FXDEFMAP(FXMenuCaption) FXMenuCaptionMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXMenuCaption::onPaint),
  FXMAPFUNC(SEL_UPDATE,0,FXMenuCaption::onUpdate),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_HELP,FXMenuCaption::onQueryHelp),
  };


// Object implementation
FXIMPLEMENT(FXMenuCaption,FXFrame,FXMenuCaptionMap,ARRAYNUMBER(FXMenuCaptionMap))



// Menu entry
FXMenuCaption::FXMenuCaption(FXComposite* p,const FXString& text,FXIcon* ic,FXuint opts):
  FXFrame(p,opts,0,0,0,0){
  label=text.extract(0,'\t','&');
  accel=text.extract(1,'\t');
  help=text.extract(2,'\t');
  icon=ic;
  font=getApp()->normalFont;
  hotkey=fxparsehotkey(text.text());
  hotoff=fxfindhotkeyoffset(text.text());
  addHotKey(hotkey);
  textColor=getApp()->foreColor;
  seltextColor=getApp()->selforeColor;
  selbackColor=getApp()->selbackColor;
  }


// Create X Window
void FXMenuCaption::create(){
  FXFrame::create();
  font->create();
  if(icon) icon->create();
  }

  
// Get default width
FXint FXMenuCaption::getDefaultWidth(){
  FXint tw,aw,iw,sw;
  tw=aw=iw=sw=0;
  if(label.text()) tw=font->getTextWidth(label.text(),label.length());
  if(accel.text()) aw=font->getTextWidth(accel.text(),accel.length());
  if(icon) iw=icon->getWidth()+5;
  if(aw && tw) sw=5;
  return FXMAX(iw,LEADSPACE)+tw+sw+aw+TRAILSPACE;
  }


// Get default height
FXint FXMenuCaption::getDefaultHeight(){
  FXint th,ih,h;
  th=ih=h=0;
  if(label.text() || accel.text()) th=font->getFontHeight()+5;
  if(icon) ih=icon->getHeight()+5;
  h=FXMAX(th,ih);
  return h;
  }


// We were asked about status text
long FXMenuCaption::onQueryHelp(FXObject* sender,FXSelector,void*){
  if(help.text() && (flags&FLAG_HELP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&help);
    return 1;
    }
  return 0;
  }


// Implement auto-hide or auto-gray modes
long FXMenuCaption::onUpdate(FXObject* sender,FXSelector sel,void* ptr){
  if(!FXFrame::onUpdate(sender,sel,ptr)){
    if(options&MENU_AUTOHIDE){if(shown()){hide();recalc();}}
    if(options&MENU_AUTOGRAY){disable();}
    }
  return 1;
  }



// Handle repaint 
long FXMenuCaption::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  FXint yy;
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  dc->setForeground(backColor);
  dc->fillRectangle(0,0,width,height);
  if(icon){
    dc->drawIcon(icon,3,(height-icon->getHeight())/2);
    }
  if(label.text()){
    dc->setTextFont(font);
    dc->setForeground(textColor);
    yy=font->getFontAscent()+(height-font->getFontHeight())/2;
    dc->drawText(LEADSPACE,yy,label.text(),label.length());
    if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
    if(0<=hotoff){
      dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
      }
    }
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Change accelerator text
void FXMenuCaption::setAccelText(const FXString& text){
  accel=text;
  }


// Change help text
void FXMenuCaption::setHelpText(const FXString& text){
  help=text;
  }


// Change text
void FXMenuCaption::setText(const FXString& text){
  if(label!=text){
    label=text.extract(0,'\t','&');
    remHotKey(hotkey);
    hotkey=fxparsehotkey(text.text());
    hotoff=fxfindhotkeyoffset(text.text());
    addHotKey(hotkey);
    recalc();
    update(0,0,width,height);
    }
  }


// Change icon
void FXMenuCaption::setIcon(FXIcon* ic){
  if(icon!=ic){
    icon=ic;
    recalc();
    update(0,0,width,height);
    }
  }


// Change font
void FXMenuCaption::setFont(FXFont* fnt){
  if(!fnt){ fxerror("%s::setFont: NULL font specified.\n",getClassName()); }
  font=fnt;
  recalc();
  update(0,0,width,height);
  }


// Set text color
void FXMenuCaption::setTextColor(FXColor clr){
  textColor=clr;
  update(0,0,width,height);
  }


// Set select background color
void FXMenuCaption::setSelBackColor(FXColor clr){
  selbackColor=clr;
  update(0,0,width,height);
  }


// Set selected text color
void FXMenuCaption::setSelTextColor(FXColor clr){
  seltextColor=clr;
  update(0,0,width,height);
  }


// Save object to stream
void FXMenuCaption::save(FXStream& store) const {
  FXFrame::save(store);
  store << label;
  store << accel;
  store << help;
  store << icon;
  store << font;
  store << hotoff;
  store << hotkey;
  store << textColor;
  store << selbackColor;
  store << seltextColor;
  }


// Load object from stream
void FXMenuCaption::load(FXStream& store){
  FXFrame::load(store);
  store >> label;
  store >> accel;
  store >> help;
  store >> icon;
  store >> font;
  store >> hotoff;
  store >> hotkey;
  store >> textColor;
  store >> selbackColor;
  store >> seltextColor;
  }  


// Zap it
FXMenuCaption::~FXMenuCaption(){
  remHotKey(hotkey);
  font=(FXFont*)-1;
  icon=(FXIcon*)-1;
  }


/*******************************************************************************/

// Map
FXDEFMAP(FXMenuCommand) FXMenuCommandMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXMenuCommand::onPaint),
  FXMAPFUNC(SEL_ENTER,0,FXMenuCommand::onEnter),
  FXMAPFUNC(SEL_LEAVE,0,FXMenuCommand::onLeave),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXMenuCommand::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXMenuCommand::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXMenuCommand::onDefault),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXMenuCommand::onDefault),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXMenuCommand::onDefault),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXMenuCommand::onDefault),
  FXMAPFUNC(SEL_ACTIVATE,0,FXMenuCommand::onActivate),
  FXMAPFUNC(SEL_DEACTIVATE,0,FXMenuCommand::onDeactivate),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_CHECK,FXMenuCommand::onCheck),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_UNCHECK,FXMenuCommand::onUncheck),
  };


// Object implementation
FXIMPLEMENT(FXMenuCommand,FXMenuCaption,FXMenuCommandMap,ARRAYNUMBER(FXMenuCommandMap))

  

// Command menu item
FXMenuCommand::FXMenuCommand(FXComposite* p,const FXString& text,FXIcon* ic,FXObject* tgt,FXSelector sel,FXuint opts):
  FXMenuCaption(p,text,ic,opts){
  flags|=FLAG_ENABLED;
  defaultCursor=getApp()->rarrowCursor;
  target=tgt;
  message=sel;
  }


// Create X window
void FXMenuCommand::create(){
  FXMenuCaption::create();
  }


// If window can have focus
FXbool FXMenuCommand::canFocus() const { 
  return 1; 
  }


// Check the menu button
long FXMenuCommand::onCheck(FXObject*,FXSelector,void*){ 
  check(); 
  return 1; 
  }


// Check the menu button
long FXMenuCommand::onUncheck(FXObject*,FXSelector,void*){ 
  uncheck(); 
  return 1; 
  }


// Pressed left button
long FXMenuCommand::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    return handle(this,MKUINT(0,SEL_ACTIVATE),ptr);
    }
  return 0;
  }


// Released left button
long FXMenuCommand::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    return handle(this,MKUINT(0,SEL_DEACTIVATE),ptr);
    }
  return 0;
  }


// Activate
long FXMenuCommand::onActivate(FXObject*,FXSelector,void*){
  return 1;
  }


// Deactivate
long FXMenuCommand::onDeactivate(FXObject* sender,FXSelector,void* ptr){
  FXbool doit=isActive();
  getParent()->handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),NULL);     // Forces popdown!
  if(isEnabled()){
    return doit && target && target->handle(this,MKUINT(message,SEL_COMMAND),(void*)1);
    }
  return 1;
  }


// Into focus chain
void FXMenuCommand::setFocus(){
  FXMenuCaption::setFocus();
  flags|=FLAG_ACTIVE;
  flags&=~FLAG_UPDATE;
  update(0,0,width,height);
  }


// Out of focus chain
void FXMenuCommand::killFocus(){
  FXMenuCaption::killFocus();
  flags&=~FLAG_ACTIVE;
  flags|=FLAG_UPDATE;
  update(0,0,width,height);
  }


// Enter
long FXMenuCommand::onEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onEnter(sender,sel,ptr);
  if(isEnabled() && canFocus()) setFocus();
  return 1;
  }


// Leave
long FXMenuCommand::onLeave(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onLeave(sender,sel,ptr);
  if(isEnabled() && canFocus()) killFocus();
  return 1;
  }


// Show as default
void FXMenuCommand::setDefault(){
  if(!(options&MENU_DEFAULT)){
    options|=MENU_DEFAULT;
    update(0,0,width,height);
    }
  }
    

// Show as other
void FXMenuCommand::setOther(){
  if(options&MENU_DEFAULT){
    options&=~MENU_DEFAULT;
    update(0,0,width,height);
    }
  }


// See if its default
FXint FXMenuCommand::isDefault() const {
  return (options&MENU_DEFAULT)!=0; 
  }


// Enable the menu entry
void FXMenuCommand::enable(){
  if(!(flags&FLAG_ENABLED)){
    FXMenuCaption::enable();
    update(0,0,width,height);
    }
  }


// Disable the menu entry
void FXMenuCommand::disable(){
  if(flags&FLAG_ENABLED){
    FXMenuCaption::disable();
    update(0,0,width,height);
    }
  }


// Show checked
void FXMenuCommand::check(){
  if(!(flags&FLAG_CHECKED)){
    flags|=FLAG_CHECKED;
    update(0,0,width,height);
    }
  }
    

// Show unchecked
void FXMenuCommand::uncheck(){
  if(flags&FLAG_CHECKED){
    flags&=~FLAG_CHECKED;
    update(0,0,width,height);
    }
  }


// Check if checked
FXint FXMenuCommand::isChecked() const {
  return (flags&FLAG_CHECKED)!=0; 
  }


// Show radio checked
void FXMenuCommand::checkRadio(){
  if(!(flags&FLAG_RCHECKED)){
    flags|=FLAG_RCHECKED;
    update(0,0,width,height);
    }
  }


// Show radio unchecked
void FXMenuCommand::uncheckRadio(){
  if(flags&FLAG_RCHECKED){
    flags&=~FLAG_RCHECKED;
    update(0,0,width,height);
    }
  }


// See if radio checked
FXint FXMenuCommand::isRadioChecked() const {
  return (flags&FLAG_RCHECKED)!=0; 
  }


// Handle repaint 
long FXMenuCommand::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  FXint yy;
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  if(isActive()){
    dc->setForeground(selbackColor);
    dc->fillRectangle(1,1,width-2,height-2);
    if(icon){
      dc->drawIcon(icon,3,(height-icon->getHeight())/2);
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(isEnabled() ? seltextColor : shadowColor);
      dc->drawText(LEADSPACE,yy,label.text(),label.length());
      if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
      if(options&MENU_DEFAULT){
        dc->drawText(LEADSPACE+1,yy,label.text(),label.length());
        if(accel.text()) dc->drawText(width-TRAILSPACE-1-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
        }
      if(0<=hotoff){
        dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
        }
      }
    }
  else if(isEnabled()){
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    if(icon){
      dc->drawIcon(icon,3,(height-icon->getHeight())/2);
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(textColor);
      dc->drawText(LEADSPACE,yy,label.text(),label.length());
      if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
      if(options&MENU_DEFAULT){
        dc->drawText(LEADSPACE+1,yy,label.text(),label.length());
        if(accel.text()) dc->drawText(width-TRAILSPACE-1-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
        }
      if(0<=hotoff){
        dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
        }
      }
    }
  else{
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    if(icon){
      dc->drawIconShaded(icon,3,(height-icon->getHeight())/2);
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(hiliteColor);
      dc->drawText(LEADSPACE+1,yy+1,label.text(),label.length());
      dc->setForeground(shadowColor);
      dc->drawText(LEADSPACE,yy,label.text(),label.length());
      if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
      if(0<=hotoff){
        dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
        }
      }
    }
  if(flags&(FLAG_CHECKED|FLAG_RCHECKED)){
    dc->setForeground(isActive() ? seltextColor : textColor);
    if(flags&FLAG_CHECKED) drawCheck(5,(height-8)/2);
    if(flags&FLAG_RCHECKED) drawBullit(7,(height-5)/2);
    }
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Draw check mark
void FXMenuCommand::drawCheck(FXint x,FXint y){
  FXSegment seg[6];
  FXDC *dc=DC();
  seg[0].x1=1+x; seg[0].y1=3+y; seg[0].x2=3+x; seg[0].y2=5+y;
  seg[1].x1=1+x; seg[1].y1=4+y; seg[1].x2=3+x; seg[1].y2=6+y;
  seg[2].x1=1+x; seg[2].y1=5+y; seg[2].x2=3+x; seg[2].y2=7+y;
  seg[3].x1=3+x; seg[3].y1=5+y; seg[3].x2=7+x; seg[3].y2=1+y;
  seg[4].x1=3+x; seg[4].y1=6+y; seg[4].x2=7+x; seg[4].y2=2+y;
  seg[5].x1=3+x; seg[5].y1=7+y; seg[5].x2=7+x; seg[5].y2=3+y;
  dc->drawLineSegments(seg,6);
  }


// Draw bullit
void FXMenuCommand::drawBullit(FXint x,FXint y){
  FXSegment seg[5];
  FXDC *dc=DC();
  seg[0].x1= 1+x; seg[0].y1 = 0+y; seg[0].x2 = 3+x; seg[0].y2 = 0+y;
  seg[1].x1= 0+x; seg[1].y1 = 1+y; seg[1].x2 = 4+x; seg[1].y2 = 1+y;
  seg[2].x1= 0+x; seg[2].y1 = 2+y; seg[2].x2 = 4+x; seg[2].y2 = 2+y;
  seg[3].x1= 0+x; seg[3].y1 = 3+y; seg[3].x2 = 4+x; seg[3].y2 = 3+y;
  seg[4].x1= 1+x; seg[4].y1 = 4+y; seg[4].x2 = 3+x; seg[4].y2 = 4+y;
  dc->drawLineSegments(seg,5);
  } 



/*******************************************************************************/

// Map
FXDEFMAP(FXMenuCascade) FXMenuCascadeMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXMenuCascade::onPaint),
  FXMAPFUNC(SEL_ENTER,0,FXMenuCascade::onEnter),
  FXMAPFUNC(SEL_LEAVE,0,FXMenuCascade::onLeave),
  FXMAPFUNC(SEL_TIMEOUT,1,FXMenuCascade::onTimeout),
  FXMAPFUNC(SEL_FOCUS_RIGHT,0,FXMenuCascade::onFocusRight),
  FXMAPFUNC(SEL_FOCUS_LEFT,0,FXMenuCascade::onFocusLeft),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXMenuCascade::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXMenuCascade::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXMenuCascade::onDefault),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXMenuCascade::onDefault),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXMenuCascade::onDefault),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXMenuCascade::onDefault),
  FXMAPFUNC(SEL_KEYPRESS,0,FXMenuCascade::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXMenuCascade::onKeyRelease),
  FXMAPFUNC(SEL_ACTIVATE,0,FXMenuCascade::onActivate),
  FXMAPFUNC(SEL_DEACTIVATE,0,FXMenuCascade::onDeactivate),
  };


// Object implementation
FXIMPLEMENT(FXMenuCascade,FXMenuCaption,FXMenuCascadeMap,ARRAYNUMBER(FXMenuCascadeMap))


// Make cascade menu button
FXMenuCascade::FXMenuCascade(FXComposite* p,const FXString& text,FXIcon* ic,FXMenuPane* pup,FXuint opts):
  FXMenuCaption(p,text,ic,opts){
  defaultCursor=getApp()->rarrowCursor;
  flags|=FLAG_ENABLED;
  pane=pup;
  timer=NULL;
  }


// Create X window; make sure pane is created
void FXMenuCascade::create(){
  FXMenuCaption::create();
  if(pane) pane->create();
  }


// Destroy X window
void FXMenuCascade::destroy(){
  if(timer){ getApp()->removeTimeout(timer); timer=NULL; }
  FXMenuCaption::destroy();
  }


// If window can have focus
FXbool FXMenuCascade::canFocus() const { 
  return 1; 
  }


// Focus moved to right:- this shows the submenu!
long FXMenuCascade::onFocusRight(FXObject*,FXSelector,void*){
  if(pane && !pane->shown()){
    FXint x,y;
    if(timer){ getApp()->removeTimeout(timer); timer=NULL; }
    translateCoordinatesTo(x,y,getRoot(),width,0);
    pane->popup(((FXMenuPane*)getParent())->getOwner(),x,y);
    }
  return 1;
  }


// Focus moved to left:- this hides the submenu!
long FXMenuCascade::onFocusLeft(FXObject*,FXSelector,void*){
  if(pane && pane->shown()){
    if(timer){ getApp()->removeTimeout(timer); timer=NULL; }
    pane->popdown();
    return 1;
    }
  return 0;
  }


// When pressed, perform ungrab, then process normally
long FXMenuCascade::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    ungrab();
    return handle(this,MKUINT(0,SEL_ACTIVATE),ptr);
    }
  return 0;
  }


// Released left button
long FXMenuCascade::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    return handle(this,MKUINT(0,SEL_DEACTIVATE),ptr);
    }
  return 0;
  }


// Activate; toggles menu
long FXMenuCascade::onActivate(FXObject*,FXSelector,void*){
  if(isEnabled()){
    if(pane){
      if(!pane->shown()){
        FXint x,y;
        if(timer){ getApp()->removeTimeout(timer); timer=NULL; }
        translateCoordinatesTo(x,y,getRoot(),width,0);
        pane->popup(((FXMenuPane*)getParent())->getOwner(),x,y);
        }
      else{
        pane->popdown();
        }
      }
    }
  return 1;
  }


// Deactivate
long FXMenuCascade::onDeactivate(FXObject*,FXSelector,void*){
  return 1;
  }


// Keyboard press; forward to submenu pane
long FXMenuCascade::onKeyPress(FXObject* sender,FXSelector sel,void* ptr){
  if(pane && pane->shown() && pane->handle(pane,sel,ptr)) return 1;
  return FXMenuCaption::onKeyPress(sender,sel,ptr);
  }


// Keyboard release; forward to submenu pane
long FXMenuCascade::onKeyRelease(FXObject* sender,FXSelector sel,void* ptr){
  if(pane && pane->shown() && pane->handle(pane,sel,ptr)) return 1;
  return FXMenuCaption::onKeyRelease(sender,sel,ptr);
  }


// Into focus chain
void FXMenuCascade::setFocus(){
  FXMenuCaption::setFocus();
  flags|=FLAG_ACTIVE;
  flags&=~FLAG_UPDATE;
  update(0,0,width,height);
  }


// Out of focus chain; hide submenu if it was up
void FXMenuCascade::killFocus(){
  FXMenuCaption::killFocus();
  if(timer){ getApp()->removeTimeout(timer); timer=NULL; }
  if(pane && pane->shown()) pane->popdown();
  flags&=~FLAG_ACTIVE;
  flags|=FLAG_UPDATE;
  update(0,0,width,height);
  }


// Enter; set timer for delayed popup
long FXMenuCascade::onEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onEnter(sender,sel,ptr);
  if(isEnabled() && canFocus()){
    if(!timer){ timer=getApp()->addTimeout(getApp()->menuPause,this,1); }
    setFocus();
    }
  return 1;
  }


// Leave
long FXMenuCascade::onLeave(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onLeave(sender,sel,ptr);
  return 1;
  }


// Timeout; pop up the submenu
long FXMenuCascade::onTimeout(FXObject*,FXSelector,void*){
  if(pane && !pane->shown()){
    FXint x,y;
    translateCoordinatesTo(x,y,getRoot(),width,0);
    pane->popup(((FXMenuPane*)getParent())->getOwner(),x,y);
    }
  timer=NULL;
  return 1;
  }


// Handle repaint 
long FXMenuCascade::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  FXint yy;
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  if(isActive()){
    dc->setForeground(selbackColor);
    dc->fillRectangle(1,1,width-2,height-2);
    if(icon){
      dc->drawIcon(icon,3,(height-icon->getHeight())/2);
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(isEnabled() ? seltextColor : shadowColor);
      dc->drawText(LEADSPACE,yy,label.text(),label.length());
      if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
      if(options&MENU_DEFAULT){
        dc->drawText(LEADSPACE+1,1+font->getFontAscent(),label.text(),label.length());
        if(accel.text()) dc->drawText(width-TRAILSPACE-1-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
        }
      if(0<=hotoff){
        dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
        }
      }
    }
  else if(isEnabled()){
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    if(icon){
      dc->drawIcon(icon,3,(height-icon->getHeight())/2);
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(textColor);
      dc->drawText(LEADSPACE,yy,label.text(),label.length());
      if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
      if(options&MENU_DEFAULT){
        dc->drawText(LEADSPACE+1,yy,label.text(),label.length());
        if(accel.text()) dc->drawText(width-TRAILSPACE-1-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
        }
      if(0<=hotoff){
        dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
        }
      }
    }
  else{
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    if(icon){
      dc->drawIconShaded(icon,3,(height-icon->getHeight())/2);
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(hiliteColor);
      dc->drawText(LEADSPACE+1,yy+1,label.text(),label.length());
      dc->setForeground(shadowColor);
      dc->drawText(LEADSPACE,yy,label.text(),label.length());
      if(accel.text()) dc->drawText(width-TRAILSPACE-font->getTextWidth(accel.text(),accel.length()),yy,accel.text(),accel.length());
      if(0<=hotoff){
        dc->drawLine(LEADSPACE+1+font->getTextWidth(label.text(),hotoff),yy+1,LEADSPACE+font->getTextWidth(label.text(),hotoff+1),yy+1);
        }
      }
    }
  dc->setForeground(isActive() ? seltextColor : textColor);
  yy=(height-8)/2;
  drawTriangle(width-TRAILSPACE+4,yy,width-TRAILSPACE+4+6,yy+8);
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Draw triangle
void FXMenuCascade::drawTriangle(FXint l,FXint t,FXint r,FXint b){
  FXPoint points[3];
  int m=(t+b)/2;
  FXDC *dc=DC();
  points[0].x=l;
  points[0].y=t;
  points[1].x=l;
  points[1].y=b;
  points[2].x=r;
  points[2].y=m;
  dc->fillPolygon(points,3);
  }


// Test if logically inside
FXbool FXMenuCascade::contains(FXint parentx,FXint parenty) const {
  FXint x,y;
  if(FXMenuCaption::contains(parentx,parenty)) return 1;
  if(getPopup() && getPopup()->shown()){
    getParent()->translateCoordinatesTo(x,y,getRoot(),parentx,parenty);
    if(getPopup()->contains(x,y)) return 1;
    }
  return 0;
  }


// Save object to stream
void FXMenuCascade::save(FXStream& store) const {
  FXMenuCaption::save(store);
  store << pane;
  }


// Load object from stream
void FXMenuCascade::load(FXStream& store){
  FXMenuCaption::load(store);
  store >> pane;
  }  


// Delete it
FXMenuCascade::~FXMenuCascade(){
  pane=(FXMenuPane*)-1;
  timer=(FXTimer*)-1;
  }


/*******************************************************************************/

// Map
FXDEFMAP(FXMenuTitle) FXMenuTitleMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXMenuTitle::onPaint),
  FXMAPFUNC(SEL_ENTER,0,FXMenuTitle::onEnter),
  FXMAPFUNC(SEL_LEAVE,0,FXMenuTitle::onLeave),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXMenuTitle::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXMenuTitle::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXMenuTitle::onDefault),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXMenuTitle::onDefault),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXMenuTitle::onDefault),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXMenuTitle::onDefault),
  FXMAPFUNC(SEL_KEYPRESS,0,FXMenuTitle::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXMenuTitle::onKeyRelease),
  FXMAPFUNC(SEL_ACTIVATE,0,FXMenuTitle::onActivate),
  FXMAPFUNC(SEL_DEACTIVATE,0,FXMenuTitle::onDeactivate),
  FXMAPFUNC(SEL_FOCUS_UP,0,FXMenuTitle::onFocusUp),
  FXMAPFUNC(SEL_FOCUS_DOWN,0,FXMenuTitle::onFocusDown),
  FXMAPFUNC(SEL_FOCUSIN,0,FXMenuTitle::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXMenuTitle::onFocusOut),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_POST,FXMenuTitle::onCmdPost),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_UNPOST,FXMenuTitle::onCmdUnpost),
  };

  
// Object implementation
FXIMPLEMENT(FXMenuTitle,FXMenuCaption,FXMenuTitleMap,ARRAYNUMBER(FXMenuTitleMap))



// Make a menu title button
FXMenuTitle::FXMenuTitle(FXComposite* p,const FXString& text,FXIcon* ic,FXMenuPane* pup,FXuint opts):
  FXMenuCaption(p,text,ic,opts){
  flags|=FLAG_ENABLED;
  options&=~LAYOUT_FILL_X;
  pane=pup;
  }


// Create X window; make sure pane is created
void FXMenuTitle::create(){
  FXMenuCaption::create();
  if(pane) pane->create();
  }


// Get default width
FXint FXMenuTitle::getDefaultWidth(){
  FXint tw,iw,w;
  tw=iw=w=0;
  if(label.text()) tw=font->getTextWidth(label.text(),label.length());
  if(icon) iw=icon->getWidth()+5;
  w=tw+iw;
  if(iw && tw) w+=5;
  return 12+w;
  }


// Get default height
FXint FXMenuTitle::getDefaultHeight(){
  FXint th,ih;
  th=ih=0;
  if(label.text()) th=font->getFontHeight()+5;
  if(icon) ih=icon->getHeight()+5;
  return FXMAX(th,ih);
  }


// If window can have focus
FXbool FXMenuTitle::canFocus() const { 
  return 1; 
  }


// When pressed, perform ungrab, then process normally
long FXMenuTitle::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
//    ungrab();
    return handle(this,MKUINT(0,SEL_ACTIVATE),ptr);
    }
  return 0;
  }


// Released left button
long FXMenuTitle::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
//    ungrab();
    return handle(this,MKUINT(0,SEL_DEACTIVATE),ptr);
    }
  return 0;
  }


// Activate
long FXMenuTitle::onActivate(FXObject*,FXSelector,void* ptr){
  if(flags&FLAG_ACTIVE){
    handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),NULL);
    }
  else{
    handle(this,MKUINT(ID_POST,SEL_COMMAND),NULL);
    }
  return 1;
  }


// Deactivate
long FXMenuTitle::onDeactivate(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
#ifdef FX_NATIVE_WIN32
  if(FXABS(ev->win_x-ev->click_x)<getApp()->dragDelta && FXABS(ev->win_y-ev->click_y)<getApp()->dragDelta) ev->moved=0;
#endif
  if(ev->moved){
    handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),ptr);
    }
  return 1;
  }


// Post the menu
long FXMenuTitle::onCmdPost(FXObject*,FXSelector,void*){
  FXint x,y;
  if(pane && !pane->shown()){
    translateCoordinatesTo(x,y,getRoot(),0,0);
    pane->popup(getParent(),x-1,y+height);
    if(!getParent()->grabbed()) getParent()->grab();
    }
  flags|=FLAG_ACTIVE;
  update(0,0,width,height);
  return 1;
  }


// Unpost the menu
long FXMenuTitle::onCmdUnpost(FXObject*,FXSelector,void*){
  if(pane && pane->shown()){
    pane->popdown();
    if(getParent()->grabbed()) getParent()->ungrab();
    }
  flags&=~FLAG_ACTIVE;
  update(0,0,width,height);
  return 1;
  }


// Focus moved down
long FXMenuTitle::onFocusDown(FXObject*,FXSelector,void*){
  if(pane && !pane->shown()){
    handle(this,MKUINT(ID_POST,SEL_COMMAND),NULL);
    return 1;
    }
  return 0;
  }


// Focus moved up
long FXMenuTitle::onFocusUp(FXObject*,FXSelector,void*){
  if(pane && pane->shown()){
    handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),NULL);
    return 1;
    }
  return 0;
  }


// Keyboard press; forward to menu pane
long FXMenuTitle::onKeyPress(FXObject* sender,FXSelector sel,void* ptr){
  if(pane && pane->shown() && pane->handle(pane,sel,ptr)) return 1;
  return FXMenuCaption::onKeyPress(sender,sel,ptr);
  }


// Keyboard release; forward to menu pane
long FXMenuTitle::onKeyRelease(FXObject* sender,FXSelector sel,void* ptr){
  if(pane && pane->shown() && pane->handle(pane,sel,ptr)) return 1;
  return FXMenuCaption::onKeyRelease(sender,sel,ptr);
  }


// Into focus chain
void FXMenuTitle::setFocus(){
  FXWindow *menuitem=getParent()->getFocus();
  FXbool active=menuitem && menuitem->isActive();
  FXMenuCaption::setFocus();
  if(active) handle(this,MKUINT(ID_POST,SEL_COMMAND),NULL);
  }
    


// Out of focus chain
void FXMenuTitle::killFocus(){
  FXMenuCaption::killFocus();
  handle(this,MKUINT(ID_UNPOST,SEL_COMMAND),NULL);
  }


// Enter
long FXMenuTitle::onEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onEnter(sender,sel,ptr);
  if(isEnabled() && canFocus() && getParent()->getFocus()) setFocus();
  return 1;
  }


// Leave
long FXMenuTitle::onLeave(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onLeave(sender,sel,ptr);
  return 1;
  }


// Handle repaint 
long FXMenuTitle::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  FXint xx,yy;
  dc->begin(this);
  dc->setClipRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  if(isActive()){
    dc->setForeground(selbackColor);
    dc->fillRectangle(0,0,width,height);
    xx=6;
    if(icon){
      dc->drawIcon(icon,5,(height-icon->getHeight())/2);
      xx=5+5+icon->getWidth();
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(seltextColor);
      dc->drawText(xx,yy,label.text(),label.length());
      if(0<=hotoff) dc->drawLine(xx+font->getTextWidth(label.text(),hotoff),yy+1,xx+font->getTextWidth(label.text(),hotoff+1)-1,yy+1);
      }
    }
  else if(isEnabled()){
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    xx=6;
    if(icon){
      dc->drawIcon(icon,5,(height-icon->getHeight())/2);
      xx=5+5+icon->getWidth();
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(textColor);
      dc->drawText(xx,yy,label.text(),label.length());
      if(0<=hotoff) dc->drawLine(xx+font->getTextWidth(label.text(),hotoff),yy+1,xx+font->getTextWidth(label.text(),hotoff+1)-1,yy+1);
      }
    }
  else{
    dc->setForeground(backColor);
    dc->fillRectangle(0,0,width,height);
    xx=6;
    if(icon){
      dc->drawIconShaded(icon,5,(height-icon->getHeight())/2);
      xx=5+5+icon->getWidth();
      }
    if(label.text()){
      yy=font->getFontAscent()+(height-font->getFontHeight())/2;
      dc->setTextFont(font);
      dc->setForeground(hiliteColor);
      dc->drawText(xx+1,yy-1,label.text(),label.length());
      dc->setForeground(shadowColor);
      dc->drawText(xx,yy,label.text(),label.length());
      if(0<=hotoff) dc->drawLine(xx+font->getTextWidth(label.text(),hotoff),yy+1,xx+font->getTextWidth(label.text(),hotoff+1)-1,yy+1);
      }
    }
  dc->clearClipRectangle();
  dc->end();
  return 1;
  }


// Gained focus
long FXMenuTitle::onFocusIn(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onFocusIn(sender,sel,ptr);
  update(0,0,width,height);
  return 1;
  }

  
// Lost focus
long FXMenuTitle::onFocusOut(FXObject* sender,FXSelector sel,void* ptr){
  FXMenuCaption::onFocusOut(sender,sel,ptr);
  update(0,0,width,height);
  return 1;
  }


// Test if logically inside
FXbool FXMenuTitle::contains(FXint parentx,FXint parenty) const {
  FXint x,y;
  if(FXMenuCaption::contains(parentx,parenty)) return 1;
  if(getPopup() && getPopup()->shown()){
    getParent()->translateCoordinatesTo(x,y,getRoot(),parentx,parenty);
    if(getPopup()->contains(x,y)) return 1;
    }
  return 0;
  }


// Save object to stream
void FXMenuTitle::save(FXStream& store) const {
  FXMenuCaption::save(store);
  store << pane;
  }


// Load object from stream
void FXMenuTitle::load(FXStream& store){
  FXMenuCaption::load(store);
  store >> pane;
  }  


// Delete it
FXMenuTitle::~FXMenuTitle(){
  pane=(FXMenuPane*)-1;
  }


/*******************************************************************************/


// Object implementation
FXIMPLEMENT(FXMenuPane,FXPopup,NULL,0)



// Build empty one
FXMenuPane::FXMenuPane(FXApp* a,FXuint opts):
  FXPopup(a,NULL,opts|FRAME_RAISED|FRAME_THICK){
  accelTable=new FXAccelTable;
  }


// Cursor is considered inside when it's in this window, or in any subwindow 
// that's open; we'll find the latter through the cascade menu, by asking it for 
// it's popup window.
FXbool FXMenuPane::contains(FXint parentx,FXint parenty) const {
  FXint x,y;
  if(FXPopup::contains(parentx,parenty)) return 1;
  if(getFocus()){
    getParent()->translateCoordinatesTo(x,y,this,parentx,parenty);
    if(getFocus()->contains(x,y)) return 1;
    }
  return 0;
  }

