/********************************************************************************
*                                                                               *
*      D e v i c e   C o n t e x t   F o r   W i n d o w - S u r f a c e s      *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* Major Contributions for Windows NT by Lyle Johnson                            *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXWDC.cpp,v 1.24 1999/01/13 11:14:04 jlj Exp $                        *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXFont.h"
#include "FXCursor.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXBitmap.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXRootWindow.h"
#include "FXShell.h"
#include "FXDC.h"
#include "FXWDC.h"


/*
  To do:
  
  - FXWDC stands for ``Window DC,'' i.e. a device context associated with
    a surface in the X-Server (Drawable, i.e. Window, or Pixmap).
    
  - There will be other types of Device Contexts, for example, a PostScript
    DC, a printer DC, and perhaps a few others [I can't think of right now].

  - For the Win32 version, could replace the code in drawLines() with
    a call to Polyline() if we also typedef'd FXPoint to POINT in fxdefs.h.
    But then the size of FXPoint is no longer platform-independent...
    JVZ: That might be acceptable, FXPoints have NO serialization anyway...
    
  - Associate a DC with a surface before you begin using it:
  
     long SomeWidget::onPaint(FXObject*,FXSelector,void* ptr){
       FXDC *dc=DC();
       dc->begin(this);
       dc->drawLine(...);
       ... jadajadajada ...
       dc->end(this);
       return 1;
       }

    The association is automatically broken when you go out of scope; the
    destructor of the FXWDC does this for you.

  - For the Win32 version, could replace the code in drawLines() with
    a call to Polyline() if we also typedef'd FXPoint to POINT in fxdefs.h.
    But then the size of FXPoint is no longer platform-independent...
      
  - Optimizations: only perform style/attribute changes just before an actual
    drawing command takes place:- X-Windows apparently already does this;
    MS-Windows also?
      
  - We assume the following initial state:
      
    BLIT Function:        BLT_SRC
    Foreground:           black (0)
    Background:           white (1)
    Line Width:           0 (meaning thinnest/fastest, no guaranteed pixelation)
    Cap Style:            CAP_BUTT
    Join Style:           JOIN_MITER
    Line Style:           LINE_SOLID
    Fill Style:           FILL_SOLID
    Fill Rule:            RULE_EVEN_ODD
    Font:                 FOX Default Font
    Other Paremeters:     To Be Determined
    
  - Under X-Windows, end() will restore the GC to the state above; flags 
    keeps track of which changes have been made to minimize the necessary 
    updating.
    
  - Under MS-Windows, extra variables keep track of old HPEN, HBRUSH, etc, and flags
    keeps track of which of these to select back into the HDC.

  - Under X, graphics_exposures should be OFF:- at least some SGI IRIX machines
    have broken implementations of graphics_exposures.
*/


/*******************************************************************************/

// Object implementation
FXIMPLEMENT(FXWDC,FXDC,NULL,0)


/* Colors in decreasing order of importance */
const FXuchar palette[][3]={
  {  0,  0,  0},
  {255,255,255},
  {  0,  0,255},
  {  0,255,  0},
  {  0,255,255},
  {255,  0,  0},
  {255,255,  0},
  {255,  0,255},
  {  0,  0,102},
  {255,204,153},
  {  0,102,  0},
  {  0,102,102},
  {  0,102,204},
  {255,153,255},
  {  0,204,102},
  {102,  0,  0},
  {102,  0,102},
  {255,153, 51},
  {102,  0,204},
  {102,102,  0},
  {102,102,102},
  {255,102,153},
  {102,102,204},
  {102,204,  0},
  {102,204,102},
  {102,204,204},
  {204,  0,102},
  {255,255,204},
  {255, 51,102},
  {255, 51, 51},
  {255, 51,  0},
  {255,204,204},
  {255,  0,204},
  {255,  0,153},
  {255,  0,102},
  {255,  0, 51},
  {255,204,255},
  {204,255,255},
  {204,255,204},
  {204,255,153},
  {204,255,102},
  {204,255, 51},
  {204,255,  0},
  {204,204,255},
  {204,204,204},
  {204,204,153},
  {204,204,102},
  {204,204, 51},
  {204,204,  0},
  {204,153,255},
  {204,153,204},
  {204,153,153},
  {204,153,102},
  {204,153, 51},
  {204,153,  0},
  {204,102,255},
  {204,102,204},
  {204,102,153},
  {204,102,102},
  {204,102, 51},
  {204,102,  0},
  {204, 51,255},
  {204, 51,204},
  {204, 51,153},
  {204, 51,102},
  {204, 51, 51},
  {204, 51,  0},
  {204,  0,255},
  {204,  0,204},
  {204,  0,153},
  {255, 51,153},
  {204,  0, 51},
  {204,  0,  0},
  {153,255,255},
  {153,255,204},
  {153,255,153},
  {153,255,102},
  {153,255, 51},
  {153,255,  0},
  {153,204,255},
  {153,204,204},
  {153,204,153},
  {153,204,102},
  {153,204, 51},
  {153,204,  0},
  {153,153,255},
  {153,153,204},
  {153,153,153},
  {153,153,102},
  {153,153, 51},
  {153,153,  0},
  {153,102,255},
  {153,102,204},
  {153,102,153},
  {153,102,102},
  {153,102, 51},
  {153,102,  0},
  {153, 51,255},
  {153, 51,204},
  {153, 51,153},
  {153, 51,102},
  {153, 51, 51},
  {153, 51,  0},
  {153,  0,255},
  {153,  0,204},
  {153,  0,153},
  {153,  0,102},
  {153,  0, 51},
  {153,  0,  0},
  {102,255,255},
  {102,255,204},
  {102,255,153},
  {102,255,102},
  {102,255, 51},
  {102,255,  0},
  {102,204,255},
  {255, 51,204},
  {102,204,153},
  {255, 51,255},
  {102,204, 51},
  {255,102,  0},
  {102,153,255},
  {102,153,204},
  {102,153,153},
  {102,153,102},
  {102,153, 51},
  {102,153,  0},
  {102,102,255},
  {255,102, 51},
  {102,102,153},
  {255,102,102},
  {102,102, 51},
  {255,102,204},
  {102, 51,255},
  {102, 51,204},
  {102, 51,153},
  {102, 51,102},
  {102, 51, 51},
  {102, 51,  0},
  {102,  0,255},
  {255,102,255},
  {102,  0,153},
  {255,153,  0},
  {102,  0, 51},
  {255,153,102},
  { 51,255,255},
  { 51,255,204},
  { 51,255,153},
  { 51,255,102},
  { 51,255, 51},
  { 51,255,  0},
  { 51,204,255},
  { 51,204,204},
  { 51,204,153},
  { 51,204,102},
  { 51,204, 51},
  { 51,204,  0},
  { 51,153,255},
  { 51,153,204},
  { 51,153,153},
  { 51,153,102},
  { 51,153, 51},
  { 51,153,  0},
  { 51,102,255},
  { 51,102,204},
  { 51,102,153},
  { 51,102,102},
  { 51,102, 51},
  { 51,102,  0},
  { 51, 51,255},
  { 51, 51,204},
  { 51, 51,153},
  { 51, 51,102},
  { 51, 51, 51},
  { 51, 51,  0},
  { 51,  0,255},
  { 51,  0,204},
  { 51,  0,153},
  { 51,  0,102},
  { 51,  0, 51},
  { 51,  0,  0},
  {255,255, 51},
  {  0,255,204},
  {  0,255,153},
  {  0,255,102},
  {  0,255, 51},
  {255,255,102},
  {  0,204,255},
  {  0,204,204},
  {  0,204,153},
  {255,153,153},
  {  0,204, 51},
  {  0,204,  0},
  {  0,153,255},
  {  0,153,204},
  {  0,153,153},
  {  0,153,102},
  {  0,153, 51},
  {  0,153,  0},
  {  0,102,255},
  {255,153,204},
  {  0,102,153},
  {255,204,  0},
  {  0,102, 51},
  {255,204, 51},
  {  0, 51,255},
  {  0, 51,204},
  {  0, 51,153},
  {  0, 51,102},
  {  0, 51, 51},
  {  0, 51,  0},
  {255,255,153},
  {  0,  0,204},
  {  0,  0,153},
  {255,204,102},
  {  0,  0, 51},
  };



/*******************************************************************************/

// Here goes the X implementation...
#ifndef FX_NATIVE_WIN32


// Construct
FXWDC::FXWDC(FXApp* a):FXDC(a){
  dc=0;
  flags=0;
  }


// Destruct
FXWDC::~FXWDC(){
  XFreeGC(getApp()->display,dc);
  flags=0;
  dc=0;
  }


// Begin locks in a drawable surface
void FXWDC::begin(FXDrawable *drawable){
  if(xid){ fxerror("%s::begin: DC already connected to drawable.\n",getClassName()); }
  xid=drawable->id();
  flags=0;
  }


// End unlock the drawable surface; restore it
void FXWDC::end(){
  XGCValues gcv;
  if(!xid){ fxerror("%s::end: DC not connected to drawable.\n",getClassName()); }
  if(flags){
    if(flags&GCFunction) gcv.function=BLT_SRC;
    if(flags&GCForeground) gcv.foreground=BlackPixel(getApp()->display,DefaultScreen(getApp()->display));
    if(flags&GCBackground) gcv.background=WhitePixel(getApp()->display,DefaultScreen(getApp()->display));
    if(flags&GCLineWidth) gcv.line_width=0;
    if(flags&GCCapStyle) gcv.cap_style=CAP_BUTT;
    if(flags&GCJoinStyle) gcv.join_style=JOIN_MITER;
    if(flags&GCLineStyle) gcv.line_style=LINE_SOLID;
    if(flags&GCFillStyle) gcv.fill_style=FILL_SOLID;
    if(flags&GCFillRule) gcv.fill_rule=RULE_EVEN_ODD;
    if(flags&GCFont) gcv.font=getApp()->normalFont->id();
    if(flags&GCClipMask) gcv.clip_mask=None;
    if(flags&GCClipXOrigin) gcv.clip_x_origin=0;
    if(flags&GCClipYOrigin) gcv.clip_y_origin=0;
    if(flags&GCDashOffset) gcv.dash_offset=0;
    if(flags&GCDashList) gcv.dashes=4;
    if(flags&GCTileStipXOrigin) gcv.ts_x_origin=0;
    if(flags&GCTileStipYOrigin) gcv.ts_y_origin=0;
    if(flags&GCSubwindowMode) gcv.subwindow_mode=ClipByChildren;
    XChangeGC(getApp()->display,dc,flags,&gcv);
    flags=0;
    }
  xid=0;
  }


// Find pixel value matching color
FXPixel FXWDC::colorpixel(FXColor clr){
  Visual *visual;
  FXPixel pixel;
  FXuint graym;
  if(!getDisplay()){ fxerror("%s::colorpixel: trying to create color before opening display.\n",getClassName()); }
  visual=DefaultVisual(getApp()->display,DefaultScreen(getApp()->display)); 
  switch(visual->c_class){
    case DirectColor:
    case TrueColor:
      pixel=(visual->red_mask&((visual->red_mask*FXREDVAL(clr))/255)) | 
            (visual->green_mask&((visual->green_mask*FXGREENVAL(clr))/255)) | 
            (visual->blue_mask&((visual->blue_mask*FXBLUEVAL(clr))/255));
      break;
    case GrayScale:
    case StaticGray:
      graym=(1<<visual->bits_per_rgb)-1;
      pixel=(graym*(77*FXREDVAL(clr)+152*FXGREENVAL(clr)+28*FXBLUEVAL(clr)))/65535;
      break;
    case StaticColor:
    case PseudoColor:
      pixel=getApp()->matchColor(FXREDVAL(clr),FXGREENVAL(clr),FXBLUEVAL(clr));
      break;
    }
  return pixel;
  }

 
// Find color of pixel
FXColor FXWDC::pixelcolor(FXPixel pix){
  Visual *visual;
  FXuint r,g,b,graym;
  if(!getDisplay()){ fxerror("%s::releaseColor: trying to create color before opening display.\n",getClassName()); }
  visual=DefaultVisual(getApp()->display,DefaultScreen(getApp()->display)); 
  switch(visual->c_class){
    case DirectColor:
    case TrueColor:
      r=(255*(pix&visual->red_mask))/visual->red_mask;
      g=(255*(pix&visual->green_mask))/visual->green_mask;
      b=(255*(pix&visual->blue_mask))/visual->blue_mask;
      break;
    case GrayScale:
    case StaticGray:
      graym=(1<<visual->bits_per_rgb)-1;
      r=(255*pix)/graym;
      b=r;
      g=r;
      break;
    case StaticColor:
    case PseudoColor:
      FXASSERT(pix<getApp()->ncolors);
      FXASSERT(getApp()->palette[pix].pixel==pix);
      r=getApp()->palette[pix].red;
      g=getApp()->palette[pix].green;
      b=getApp()->palette[pix].blue;
      break;
    }
  return FXRGB(r,g,b);
  }
  
  

void FXWDC::drawPoint(FXint x,FXint y){ 
  if(!xid){ fxerror("%s::drawPoint: DC not connected to drawable.\n",getClassName()); }
  XDrawPoint(getApp()->display,xid,dc,x,y); 
  }


void FXWDC::drawPoints(const FXPoint* points,FXuint npoints){ 
  if(!xid){ fxerror("%s::drawPoints: DC not connected to drawable.\n",getClassName()); }
  XDrawPoints(getApp()->display,xid,dc,(XPoint*)points,npoints,CoordModeOrigin); 
  }


void FXWDC::drawPointsRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::drawPointsRel: DC not connected to drawable.\n",getClassName()); }
  XDrawPoints(getApp()->display,xid,dc,(XPoint*)points,npoints,CoordModePrevious); 
  }


void FXWDC::drawLine(FXint x1,FXint y1,FXint x2,FXint y2){
  if(!xid){ fxerror("%s::drawLine: DC not connected to drawable.\n",getClassName()); }
  XDrawLine(getApp()->display,xid,dc,x1,y1,x2,y2);
  }


void FXWDC::drawLines(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::drawLines: DC not connected to drawable.\n",getClassName()); }
  XDrawLines(getApp()->display,xid,dc,(XPoint*)points,npoints,CoordModeOrigin); 
  }


void FXWDC::drawLinesRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::drawLinesRel: DC not connected to drawable.\n",getClassName()); }
  XDrawLines(getApp()->display,xid,dc,(XPoint*)points,npoints,CoordModePrevious); 
  }


void FXWDC::drawLineSegments(const FXSegment* segments,FXuint nsegments){
  if(!xid){ fxerror("%s::drawLineSegments: DC not connected to drawable.\n",getClassName()); }
  XDrawSegments(getApp()->display,xid,dc,(XSegment*)segments,nsegments); 
  }
 

void FXWDC::drawRectangle(FXint x,FXint y,FXint w,FXint h){
  if(!xid){ fxerror("%s::drawRectangle: DC not connected to drawable.\n",getClassName()); }
  XDrawRectangle(getApp()->display,xid,dc,x,y,w,h);
  }


void FXWDC::drawRectangles(const FXRectangle* rectangles,FXuint nrectangles){
  if(!xid){ fxerror("%s::drawRectangles: DC not connected to drawable.\n",getClassName()); }
  XDrawRectangles(getApp()->display,xid,dc,(XRectangle*)rectangles,nrectangles);
  }


void FXWDC::drawArc(FXint x,FXint y,FXint w,FXint h,FXint ang1,FXint ang2){
  if(!xid){ fxerror("%s::drawArc: DC not connected to drawable.\n",getClassName()); }
  XDrawArc(getApp()->display,xid,dc,x,y,w,h,ang1,ang2);
  }


void FXWDC::drawArcs(const FXArc* arcs,FXuint narcs){
  if(!xid){ fxerror("%s::drawArcs: DC not connected to drawable.\n",getClassName()); }
  XDrawArcs(getApp()->display,xid,dc,(XArc*)arcs,narcs);
  }


void FXWDC::fillRectangle(FXint x,FXint y,FXint w,FXint h){
  if(!xid){ fxerror("%s::fillRectangle: DC not connected to drawable.\n",getClassName()); }
  XFillRectangle(getApp()->display,xid,dc,x,y,w,h);
  }


void FXWDC::fillRectangles(const FXRectangle* rectangles,FXuint nrectangles){
  if(!xid){ fxerror("%s::fillRectangles: DC not connected to drawable.\n",getClassName()); }
  XFillRectangles(getApp()->display,xid,dc,(XRectangle*)rectangles,nrectangles);
  }


void FXWDC::fillArc(FXint x,FXint y,FXint w,FXint h,FXint ang1,FXint ang2){
  if(!xid){ fxerror("%s::fillArc: DC not connected to drawable.\n",getClassName()); }
  XFillArc(getApp()->display,xid,dc,x,y,w,h,ang1,ang2);
  }


void FXWDC::fillArcs(const FXArc* arcs,FXuint narcs){
  if(!xid){ fxerror("%s::fillArcs: DC not connected to drawable.\n",getClassName()); }
  XFillArcs(getApp()->display,xid,dc,(XArc*)arcs,narcs);
  }


void FXWDC::fillPolygon(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillPolygon: DC not connected to drawable.\n",getClassName()); }
  XFillPolygon(getApp()->display,xid,dc,(XPoint*)points,npoints,Convex,CoordModeOrigin);
  }


void FXWDC::fillConcavePolygon(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillConcavePolygon: DC not connected to drawable.\n",getClassName()); }
  XFillPolygon(getApp()->display,xid,dc,(XPoint*)points,npoints,Nonconvex,CoordModeOrigin);
  }


void FXWDC::fillComplexPolygon(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillComplexPolygon: DC not connected to drawable.\n",getClassName()); }
  XFillPolygon(getApp()->display,xid,dc,(XPoint*)points,npoints,Complex,CoordModeOrigin);
  }


void FXWDC::fillPolygonRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillPolygonRel: DC not connected to drawable.\n",getClassName()); }
  XFillPolygon(getApp()->display,xid,dc,(XPoint*)points,npoints,Convex,CoordModePrevious);
  }


void FXWDC::fillConcavePolygonRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillConcavePolygonRel: DC not connected to drawable.\n",getClassName()); }
  XFillPolygon(getApp()->display,xid,dc,(XPoint*)points,npoints,Nonconvex,CoordModePrevious);
  }


void FXWDC::fillComplexPolygonRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillComplexPolygonRel: DC not connected to drawable.\n",getClassName()); }
  XFillPolygon(getApp()->display,xid,dc,(XPoint*)points,npoints,Complex,CoordModePrevious);
  }


void FXWDC::drawText(FXint x,FXint y,const FXchar* string,FXuint length){
  if(!xid){ fxerror("%s::drawText: DC not connected to drawable.\n",getClassName()); }
  XDrawString(getApp()->display,xid,dc,x,y,(char*)string,length);
  }


void FXWDC::drawImageText(FXint x,FXint y,const FXchar* string,FXuint length){
  if(!xid){ fxerror("%s::drawImageText: DC not connected to drawable.\n",getClassName()); }
  XDrawImageString(getApp()->display,xid,dc,x,y,(char*)string,length);
  }


void FXWDC::drawArea(const FXDrawable* source,FXint sx,FXint sy,FXint sw,FXint sh,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawArea: DC not connected to drawable.\n",getClassName()); }
  if(!source || !source->xid){ fxerror("%s::drawArea: illegal source specified.\n",getClassName()); }
  XCopyArea(getApp()->display,source->xid,xid,dc,sx,sy,sw,sh,dx,dy);
  }


void FXWDC::drawImage(const FXImage* image,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawImage: DC not connected to drawable.\n",getClassName()); }
  if(!image || !image->xid){ fxerror("%s::drawImage: illegal image specified.\n",getClassName()); }
  XCopyArea(getApp()->display,image->xid,xid,dc,0,0,image->width,image->height,dx,dy);
  }


void FXWDC::drawIcon(const FXIcon* icon,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawIcon: DC not connected to drawable.\n",getClassName()); }
  if(!icon || !icon->xid || !icon->shape){ fxerror("%s::drawIcon: illegal icon specified.\n",getClassName()); }
  XGCValues gcv;  
  gcv.clip_mask=icon->shape;
  gcv.clip_x_origin=dx;
  gcv.clip_y_origin=dy;
  XChangeGC(getApp()->display,dc,GCClipMask|GCClipXOrigin|GCClipYOrigin,&gcv);
  XCopyArea(getApp()->display,icon->xid,xid,dc,0,0,icon->width,icon->height,dx,dy);
  gcv.clip_mask=None;
  XChangeGC(getApp()->display,dc,GCClipMask,&gcv);
  }


void FXWDC::drawIconShaded(const FXIcon* icon,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawIconShaded: DC not connected to drawable.\n",getClassName()); }
  if(!icon || !icon->xid || !icon->shape){ fxerror("%s::drawIconShaded: illegal icon specified.\n",getClassName()); }
  if(!getApp()->grayBitmap || !getApp()->grayBitmap->xid){ fxerror("%s::drawIconShaded: standard gray bitmap not present.\n",getClassName()); }
  XGCValues gcv;  
  gcv.clip_mask=icon->shape;
  gcv.clip_x_origin=dx;
  gcv.clip_y_origin=dy;
  XChangeGC(getApp()->display,dc,GCClipMask|GCClipXOrigin|GCClipYOrigin,&gcv);
  XCopyArea(getApp()->display,icon->xid,xid,dc,0,0,icon->width,icon->height,dx,dy);
  gcv.function=BLT_SRC;
  gcv.stipple=getApp()->grayBitmap->xid;
  gcv.fill_style=FILL_STIPPLED;
  gcv.foreground=colorpixel(getApp()->backColor);
  XChangeGC(getApp()->display,dc,GCForeground|GCFunction|GCStipple|GCFillStyle,&gcv);
  XFillRectangle(getApp()->display,xid,dc,dx,dy,icon->width,icon->height);
  gcv.clip_mask=None;
  gcv.function=BLT_SRC;
  gcv.fill_style=FILL_SOLID;
  XChangeGC(getApp()->display,dc,GCClipMask|GCFunction|GCFillStyle,&gcv);
  }


void FXWDC::drawHashBox(FXint x,FXint y,FXint w,FXint h,FXint b){
  XGCValues gcv;  
  if(!xid){ fxerror("%s::drawHashBox: DC not connected to drawable.\n",getClassName()); }
  if(!getApp()->grayBitmap || !getApp()->grayBitmap->xid){ fxerror("%s::drawHashBox: standard gray bitmap not present.\n",getClassName()); }
  gcv.stipple=getApp()->grayBitmap->xid;
  gcv.fill_style=FILL_STIPPLED;
  XChangeGC(getApp()->display,dc,GCStipple|GCFillStyle,&gcv);
  XFillRectangle(getApp()->display,xid,dc,x,y,w-b,b);
  XFillRectangle(getApp()->display,xid,dc,x+w-b,y,b,h-b);
  XFillRectangle(getApp()->display,xid,dc,x+b,y+h-b,w-b,b);
  XFillRectangle(getApp()->display,xid,dc,x,y+b,b,h-b);
  gcv.stipple=getApp()->blackstipple;   // Needed for IRIX6.4 bug workaround!
  gcv.fill_style=FILL_SOLID;
  XChangeGC(getApp()->display,dc,GCStipple|GCFillStyle,&gcv);
  }


void FXWDC::setForeground(FXColor clr){
  if(!xid){ fxerror("%s::setForeground: DC not connected to drawable.\n",getClassName()); }
  XSetForeground(getApp()->display,dc,colorpixel(clr));
  flags|=GCForeground;
  }


void FXWDC::setBackground(FXColor clr){
  if(!xid){ fxerror("%s::setBackground: DC not connected to drawable.\n",getClassName()); }
  XSetBackground(getApp()->display,dc,colorpixel(clr));
  flags|=GCBackground;
  }


void FXWDC::setDashes(FXuint dashoffset,const FXchar *dashlist,FXuint n){
  if(!xid){ fxerror("%s::setDashes: DC not connected to drawable.\n",getClassName()); }
  XSetDashes(getApp()->display,dc,dashoffset,(char*)dashlist,n);
  flags|=(GCDashList|GCDashOffset);
  }


void FXWDC::setLineWidth(FXuint linewidth){ 
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setLineWidth: DC not connected to drawable.\n",getClassName()); }
  gcv.line_width=linewidth;
  XChangeGC(getApp()->display,dc,GCLineWidth,&gcv);
  flags|=GCLineWidth;
  }


void FXWDC::setLineCap(FXCapStyle capstyle){ 
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setLineCap: DC not connected to drawable.\n",getClassName()); }
  gcv.cap_style=capstyle;
  XChangeGC(getApp()->display,dc,GCCapStyle,&gcv);
  flags|=GCCapStyle;
  }


void FXWDC::setLineJoin(FXJoinStyle joinstyle){ 
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setLineJoin: DC not connected to drawable.\n",getClassName()); }
  gcv.join_style=joinstyle;
  XChangeGC(getApp()->display,dc,GCJoinStyle,&gcv);
  flags|=GCJoinStyle;
  }


void FXWDC::setLineStyle(FXLineStyle linestyle){ 
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setLineStyle: DC not connected to drawable.\n",getClassName()); }
  gcv.line_style=linestyle;
  XChangeGC(getApp()->display,dc,GCLineStyle,&gcv);
  flags|=GCLineStyle;
  }


void FXWDC::setFillStyle(FXFillStyle fillstyle){
  if(!xid){ fxerror("%s::setFillStyle: DC not connected to drawable.\n",getClassName()); }
  XSetFillStyle(getApp()->display,dc,fillstyle);
  flags|=GCFillStyle;
  }
  
  
void FXWDC::setFillRule(FXFillRule fillrule){
  if(!xid){ fxerror("%s::setFillRule: DC not connected to drawable.\n",getClassName()); }
  XSetFillRule(getApp()->display,dc,fillrule);
  flags|=GCFillRule;
  }


void FXWDC::setFunction(FXFunction func){
  if(!xid){ fxerror("%s::setFunction: DC not connected to drawable.\n",getClassName()); }
  XSetFunction(getApp()->display,dc,func);
  flags|=GCFunction;
  }


void FXWDC::setTile(FXImage* tile,FXint dx,FXint dy){
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setTile: DC not connected to drawable.\n",getClassName()); }
  if(!tile || !tile->xid){ fxerror("%s::setTile: illegal image specified.\n",getClassName()); }
  gcv.tile=tile->xid;
  gcv.ts_x_origin=dx;
  gcv.ts_y_origin=dy;
  XChangeGC(getApp()->display,dc,GCTileStipXOrigin|GCTileStipYOrigin|GCTile,&gcv);
  if(dx) flags|=GCTileStipXOrigin;
  if(dy) flags|=GCTileStipYOrigin;
  }


void FXWDC::setStipple(FXBitmap* stipple,FXint dx,FXint dy){
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setStipple: DC not connected to drawable.\n",getClassName()); }
  if(!stipple || !stipple->xid){ fxerror("%s::setStipple: illegal image specified.\n",getClassName()); }
  gcv.stipple=stipple->xid;
  gcv.ts_x_origin=dx;
  gcv.ts_y_origin=dy;
  XChangeGC(getApp()->display,dc,GCTileStipXOrigin|GCTileStipYOrigin|GCStipple,&gcv);
  if(dx) flags|=GCTileStipXOrigin;
  if(dy) flags|=GCTileStipYOrigin;
  }


void FXWDC::setClipRectangle(FXint x,FXint y,FXint w,FXint h){
  XRectangle c;
  if(!xid){ fxerror("%s::setClipRectangle: DC not connected to drawable.\n",getClassName()); }
  c.x=x; c.y=y; c.width=w; c.height=h;
  XSetClipRectangles(getApp()->display,dc,0,0,&c,1,Unsorted);
  flags|=GCClipMask;
  }
 

void FXWDC::clearClipRectangle(){
  if(!xid){ fxerror("%s::clearClipRectangle: DC not connected to drawable.\n",getClassName()); }
  XSetClipMask(getApp()->display,dc,None);
  flags&=~GCClipMask;
  }


void FXWDC::setClipMask(FXBitmap* mask,FXint dx,FXint dy){
  XGCValues gcv;  
  if(!xid){ fxerror("%s::setClipMask: DC not connected to drawable.\n",getClassName()); }
  if(!mask || !mask->xid){ fxerror("%s::setClipMask: illegal mask specified.\n",getClassName()); }
  gcv.clip_mask=mask->xid;
  gcv.clip_x_origin=dx;
  gcv.clip_y_origin=dy;
  XChangeGC(getApp()->display,dc,GCClipMask|GCClipXOrigin|GCClipYOrigin,&gcv);
  if(dx) flags|=GCClipXOrigin;
  if(dy) flags|=GCClipYOrigin;
  flags|=GCClipMask;
  }


void FXWDC::clearClipMask(){
  if(!xid){ fxerror("%s::clearClipMask: DC not connected to drawable.\n",getClassName()); }
  XSetClipMask(getApp()->display,dc,None);
  flags&=~GCClipMask;
  }


void FXWDC::clipChildWindows(FXbool yes){ 
  if(!xid){ fxerror("%s::clipChildWindows: DC not connected to drawable.\n",getClassName()); }
  if(yes){
    XSetSubwindowMode(getDisplay(),dc,ClipByChildren);    // This is the default!
    flags&=~GCSubwindowMode;
    }
  else{
    XSetSubwindowMode(getDisplay(),dc,IncludeInferiors);
    flags|=GCSubwindowMode;
    }
  }


void FXWDC::setTextFont(FXFont *font){
  if(!xid){ fxerror("%s::setTextFont: DC not connected to drawable.\n",getClassName()); }
  if(!font || !font->id()){ fxerror("%s::setTextFont: illegal or NULL font specified.\n",getClassName()); }
  XSetFont(getApp()->display,dc,font->id());
  flags|=GCFont;
  }


/*******************************************************************************/

// And here goes the WIN32 native implementation
#else


// Construct
FXWDC::FXWDC(FXApp* a):FXDC(a){
  dc=0;
  rop=BLT_SRC;
  flags=0;
  stipplebitmap=0;
  tileimage=0;
  dwPenType=PS_GEOMETRIC;
  dwPenStyle=PS_SOLID;
  dwEndCapAttributes=PS_ENDCAP_ROUND;
  dwJoinAttributes=PS_JOIN_BEVEL;
  }


// Destruct
FXWDC::~FXWDC(){
  if(dc){ReleaseDC((HWND)xid,dc);}
  dc=0;
  flags=0;
  }


// Begin locks in a drawable surface
void FXWDC::begin(FXDrawable *drawable){
  if(xid){ fxerror("%s::begin: DC already connected to drawable.\n.",getClassName()); }
  xid=drawable->id();
  dc=GetDC((HWND)xid);

  // Create our default pen (black, solid, one pixel wide)
  LOGBRUSH lb;
  lb.lbStyle=BS_SOLID;
  lb.lbColor=PALETTERGB(0,0,0);
  lb.lbHatch=0;
  SelectObject(dc,ExtCreatePen(dwPenType|dwPenStyle|dwEndCapAttributes|dwJoinAttributes,1,&lb,0,NULL));

  // Create our default brush (solid white, for fills)
  SelectObject(dc,CreateSolidBrush(PALETTERGB(255,255,255)));

  // Text alignment
  SetTextAlign(dc,TA_BASELINE|TA_LEFT);

  // Polygon fill mode
  SetPolyFillMode(dc,ALTERNATE);
  }


// End unlock the drawable surface 
void FXWDC::end(){
  DWORD dwStyle;
  if(!xid){ fxerror("%s::end: DC not connected to drawable.\n",getClassName()); }
  dwStyle=GetWindowLong((HWND)xid,GWL_STYLE);
  dwStyle|=WS_CLIPCHILDREN;
  SetWindowLong((HWND)xid,GWL_STYLE,dwStyle);
  DeleteObject(SelectObject(dc,GetStockObject(BLACK_PEN)));
  DeleteObject(SelectObject(dc,GetStockObject(WHITE_BRUSH)));
  ReleaseDC((HWND)xid,dc);
  dc=0;
  xid=0;
  }


// Find pixel value matching color
FXPixel FXWDC::colorpixel(FXColor clr){
  return PALETTERGB(FXREDVAL(clr),FXGREENVAL(clr),FXBLUEVAL(clr));
  }

// Find color matching pixel value
FXColor FXWDC::pixelcolor(FXPixel pix){
  return PALETTEINDEX(pix);
  }


void FXWDC::drawPoint(FXint x,FXint y){ 
  if(!xid){ fxerror("%s::drawPoint: DC not connected to drawable.\n",getClassName()); }
  // Get a handle to current pen
  HPEN hPen=(HPEN)GetCurrentObject(dc,OBJ_PEN);
  
  // Get its attributes (specifically, its color)
  EXTLOGPEN elp;
  GetObject(hPen,sizeof(EXTLOGPEN),&elp);
  SetPixel(dc,x,y,elp.elpColor);
  }


void FXWDC::drawPoints(const FXPoint* points,FXuint npoints){ 
  if(!xid){ fxerror("%s::drawPoints: DC not connected to drawable.\n",getClassName()); }
  for(FXuint i=0; i<npoints; i++) drawPoint(points[i].x,points[i].y);
  }


void FXWDC::drawPointsRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::drawPointsRel: DC not connected to drawable.\n",getClassName()); }
  int x=0,y=0;
  for(FXuint i=0; i < npoints; i++){
    x+=points[i].x;
    y+=points[i].y;
    drawPoint(x,y);
    }
  }


void FXWDC::drawLine(FXint x1,FXint y1,FXint x2,FXint y2){
  if(!xid){ fxerror("%s::drawLine: DC not connected to drawable.\n",getClassName()); }
  MoveToEx(dc,x1,y1,NULL);
  BeginPath(dc); LineTo(dc,x2,y2); EndPath(dc); StrokePath(dc);
  }


void FXWDC::drawLines(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::drawLines: DC not connected to drawable.\n",getClassName()); }
  MoveToEx(dc,points[0].x,points[0].y,NULL);
  BeginPath(dc);
  for(FXuint i=1; i < npoints; i++) LineTo(dc,points[i].x,points[i].y);
  EndPath(dc);
  StrokePath(dc);
  }


void FXWDC::drawLinesRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::drawLinesRel: DC not connected to drawable.\n",getClassName()); }
  int x=points[0].x,y=points[0].y;
  MoveToEx(dc,x,y,NULL);
  BeginPath(dc);
  for(FXuint i=1; i < npoints; i++){
    x+=points[i].x;
    y+=points[i].y;
    LineTo(dc,x,y);
    }
  EndPath(dc);
  StrokePath(dc);
  }


void FXWDC::drawLineSegments(const FXSegment* segments,FXuint nsegments){
  if(!xid){ fxerror("%s::drawLineSegments: DC not connected to drawable.\n",getClassName()); }
  BeginPath(dc);
  for(FXuint i=0; i < nsegments; i++){
    MoveToEx(dc,segments[i].x1,segments[i].y1,NULL);
    LineTo(dc,segments[i].x2,segments[i].y2);
    }
  EndPath(dc);
  StrokePath(dc);
  }
 

// Unfilled rectangle
void FXWDC::drawRectangle(FXint x,FXint y,FXint w,FXint h){
  if(!xid){ fxerror("%s::drawRectangle: DC not connected to drawable.\n",getClassName()); }
  SaveDC(dc);
  SelectObject(dc,GetStockObject(NULL_BRUSH));
  BeginPath(dc);
  Rectangle(dc,x,y,x+w+1,y+h+1);
  EndPath(dc);
  StrokePath(dc);
  RestoreDC(dc,-1);
  }


void FXWDC::drawRectangles(const FXRectangle* rectangles,FXuint nrectangles){
  if(!xid){ fxerror("%s::drawRectangles: DC not connected to drawable.\n",getClassName()); }
  SaveDC(dc);
  SelectObject(dc,GetStockObject(NULL_BRUSH));
  BeginPath(dc);
  for(FXuint i=0; i < nrectangles; i++) Rectangle(dc,rectangles[i].x,rectangles[i].y,rectangles[i].x+rectangles[i].w+1,rectangles[i].y+rectangles[i].h+1);
  EndPath(dc);
  StrokePath(dc);
  RestoreDC(dc,-1);
  }


void FXWDC::drawArc(FXint x,FXint y,FXint w,FXint h,FXint ang1,FXint ang2){
  if(!xid){ fxerror("%s::drawArc: DC not connected to drawable.\n",getClassName()); }
  ang1/=64; // Angles are in 64th of degree
  ang2/=64;
  int xStart=x+w/2+int(w*cos(ang1*PI/180.0));
  int yStart=y+h/2-int(h*sin(ang1*PI/180.0));
  int xEnd=x+w/2+int(w*cos(ang2*PI/180.0));
  int yEnd=y+h/2-int(h*sin(ang2*PI/180.0));
  BeginPath(dc);
  Arc(dc,x,y,x+w,y+w,xStart,xEnd,yStart,yEnd);
  EndPath(dc);
  StrokePath(dc);
  }


void FXWDC::drawArcs(const FXArc* arcs,FXuint narcs){
  if(!xid){ fxerror("%s::drawArcs: DC not connected to drawable.\n",getClassName()); }
  for(FXuint i=0; i<narcs; i++)
    drawArc(arcs[i].x,arcs[i].y,arcs[i].w,arcs[i].h,arcs[i].a,arcs[i].b);
  }


// Fill using currently selected ROP code
void FXWDC::fillRectangle(FXint x,FXint y,FXint w,FXint h){
  if(!xid){ fxerror("%s::fillRectangle: DC not connected to drawable.\n",getClassName()); }
  switch(rop){
    case BLT_CLR:                     // D := 0
      PatBlt(dc,x,y,w,h,BLACKNESS); 
      break;
    case BLT_NOT_DST:                 // D := ~D
      PatBlt(dc,x,y,w,h,DSTINVERT); 
      break;
    case BLT_SET:                     // D := 1
      PatBlt(dc,x,y,w,h,WHITENESS); 
      break;
    case BLT_SRC:                     // D := S
      PatBlt(dc,x,y,w,h,PATCOPY);
      break;
    case BLT_SRC_AND_DST:             // D := S & D
    case BLT_SRC_AND_NOT_DST:         // D := S & ~D
    case BLT_NOT_SRC_AND_DST:         // D := ~S & D
    case BLT_SRC_XOR_DST:             // D := S ^ D
    case BLT_SRC_OR_DST:              // D := S | D
    case BLT_NOT_SRC_AND_NOT_DST:     // D := ~S & ~D  ==  D := ~(S | D)
    case BLT_NOT_SRC_XOR_DST:         // D := ~S ^ D
    case BLT_SRC_OR_NOT_DST:          // D := S | ~D
    case BLT_NOT_SRC:                 // D := ~S
    case BLT_NOT_SRC_OR_DST:          // D := ~S | D
    case BLT_NOT_SRC_OR_NOT_DST:      // D := ~S | ~D  ==  ~(S & D)
    case BLT_DST:
      // Invalid ROP because no source
      FXASSERT(FALSE);
      break;
    default:
      FXASSERT(FALSE);
    }
  }


// Fill using currently selected ROP code
void FXWDC::fillRectangles(const FXRectangle* rectangles,FXuint nrectangles){
  if(!xid){ fxerror("%s::fillRectangles: DC not connected to drawable.\n",getClassName()); }
  for(FXuint i=0; i < nrectangles; i++) fillRectangle(rectangles[i].x,rectangles[i].y,rectangles[i].w,rectangles[i].h);
  }


// Draw filled arc
void FXWDC::fillArc(FXint x,FXint y,FXint w,FXint h,FXint ang1,FXint ang2){
  if(!xid){ fxerror("%s::fillArc: DC not connected to drawable.\n",getClassName()); }
  ang1/=64; // Angles are in 64th of degree
  ang2/=64;
  int xStart=x+w/2+int(w*cos(ang1*PI/180.0));
  int yStart=y+h/2-int(h*sin(ang1*PI/180.0));
  int xEnd=x+w/2+int(w*cos(ang2*PI/180.0));
  int yEnd=y+h/2-int(h*sin(ang2*PI/180.0));
  HPEN hPen=(HPEN)SelectObject(dc,GetStockObject(NULL_PEN));
  Pie(dc,x,y,x+w,y+h,xStart,yStart,xEnd,yEnd);
  SelectObject(dc,hPen);
  }


void FXWDC::fillArcs(const FXArc* arcs,FXuint narcs){
  if(!xid){ fxerror("%s::fillArcs: DC not connected to drawable.\n",getClassName()); }
  for(FXuint i=0; i<narcs; i++)
    fillArc(arcs[i].x,arcs[i].y,arcs[i].w,arcs[i].h,arcs[i].a,arcs[i].b);
  }


// Filled simple polygon
void FXWDC::fillPolygon(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillPolygon: DC not connected to drawable.\n",getClassName()); }
  POINT *pts=new POINT[npoints];
  for(FXuint i=0; i<npoints; i++){
    pts[i].x=points[i].x;
    pts[i].y=points[i].y;
    }
  Polygon(dc,pts,npoints);
  delete [] pts;
  }


void FXWDC::fillConcavePolygon(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillConcavePolygon: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }


void FXWDC::fillComplexPolygon(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillComplexPolygon: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }


// Filled simple polygon with relative points
void FXWDC::fillPolygonRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillPolygonRel: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }


void FXWDC::fillConcavePolygonRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillConcavePolygonRel: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }


void FXWDC::fillComplexPolygonRel(const FXPoint* points,FXuint npoints){
  if(!xid){ fxerror("%s::fillComplexPolygonRel: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }


// Draw string (only foreground bits)
void FXWDC::drawText(FXint x,FXint y,const FXchar* string,FXuint length){
  if(!xid){ fxerror("%s::drawText: DC not connected to drawable.\n",getClassName()); }
  SaveDC(dc);
  SetBkMode(dc,TRANSPARENT);
  TextOut(dc,x,y,string,length);
  RestoreDC(dc,-1);
  }


// Draw string (both foreground and background bits)
void FXWDC::drawImageText(FXint x,FXint y,const FXchar* string,FXuint length){
  if(!xid){ fxerror("%s::drawImageText: DC not connected to drawable.\n",getClassName()); }
  SaveDC(dc);
  SetBkMode(dc,OPAQUE);
  TextOut(dc,x,y,string,length);
  RestoreDC(dc,-1);
  }


// Draw area from source
// Some of these ROP codes do not have names; the full list can be found in the MSDN docs
// at Platform SDK/Reference/Appendixes/Win32 Appendixes/Raster Operation Codes/Ternary Raster Operations
void FXWDC::drawArea(const FXDrawable* source,FXint sx,FXint sy,FXint sw,FXint sh,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawArea: DC not connected to drawable.\n",getClassName()); }
  if(!source || !source->xid){ fxerror("%s::drawArea: illegal source specified.\n",getClassName()); }
  HDC shdc=GetDC((HWND)source->xid);//// Different for Bitmaps and Windows!!!!
  switch(rop){
    case BLT_CLR:                     // D := 0
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,BLACKNESS); 
      break;
    case BLT_SRC_AND_DST:             // D := S & D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,SRCAND); 
      break;
    case BLT_SRC_AND_NOT_DST:         // D := S & ~D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,SRCERASE); 
      break;
    case BLT_SRC:                     // D := S
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,SRCCOPY); 
      break;
    case BLT_NOT_SRC_AND_DST:         // D := ~S & D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,0x220326);
      break;
    case BLT_SRC_XOR_DST:             // D := S ^ D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,SRCINVERT); 
      break;
    case BLT_SRC_OR_DST:              // D := S | D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,SRCPAINT); 
      break;
    case BLT_NOT_SRC_AND_NOT_DST:     // D := ~S & ~D  ==  D := ~(S | D)
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,NOTSRCERASE); 
      break;
    case BLT_NOT_SRC_XOR_DST:         // D := ~S ^ D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,0x990066); // Not sure about this one
      break;
    case BLT_NOT_DST:                 // D := ~D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,DSTINVERT); 
      break;
    case BLT_SRC_OR_NOT_DST:          // D := S | ~D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,0xDD0228); 
      break;
    case BLT_NOT_SRC:                 // D := ~S
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,NOTSRCCOPY); 
      break;
    case BLT_NOT_SRC_OR_DST:          // D := ~S | D
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,MERGEPAINT); 
      break;
    case BLT_NOT_SRC_OR_NOT_DST:      // D := ~S | ~D  ==  ~(S & D)
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,0x7700E6); 
      break;
    case BLT_SET:                     // D := 1
      BitBlt(dc,dx,dy,sw,sh,shdc,sx,sy,WHITENESS); 
    case BLT_DST: 
      break;
    }
  ReleaseDC((HWND)source->xid,shdc);
  }


// Draw image
void FXWDC::drawImage(const FXImage* image,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawImage: DC not connected to drawable.\n",getClassName()); }
  if(!image || !image->xid){ fxerror("%s::drawImage: illegal image specified.\n",getClassName()); }
  HDC dcMem=CreateCompatibleDC(dc);
  SelectObject(dcMem,image->xid);
  SetMapMode(dcMem,GetMapMode(dc));
  POINT ptSize;
  ptSize.x=image->width;
  ptSize.y=image->height;
  DPtoLP(dc,&ptSize,1);
  POINT ptOrg;
  ptOrg.x=0; ptOrg.y=0;
  DPtoLP(dc,&ptOrg,1);
  BitBlt(dc,dx,dy,ptSize.x,ptSize.y,dcMem,ptOrg.x,ptOrg.y,SRCCOPY);
  DeleteDC(dcMem);
  }


// Draw icon
void FXWDC::drawIcon(const FXIcon* icon,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawIcon: DC not connected to drawable.\n",getClassName()); }
  if(!icon || !icon->xid || !icon->shape){ fxerror("%s::drawIcon: illegal icon specified.\n",getClassName()); }
  HDC hdcSrc=CreateCompatibleDC(dc);
  HDC hdcMsk=CreateCompatibleDC(dc);
  SelectObject(hdcSrc,icon->xid);
  SelectObject(hdcMsk,icon->shape);
  COLORREF crOldBack=SetBkColor(dc,RGB(255,255,255));
  COLORREF crOldText=SetTextColor(dc,RGB(0,0,0));
  BitBlt(dc,dx,dy,icon->width,icon->height,hdcSrc,0,0,SRCINVERT);
  BitBlt(dc,dx,dy,icon->width,icon->height,hdcMsk,0,0,SRCAND);
  BitBlt(dc,dx,dy,icon->width,icon->height,hdcSrc,0,0,SRCINVERT);
  DeleteDC(hdcMsk);
  DeleteDC(hdcSrc);
  SetBkColor(dc,crOldBack);
  SetTextColor(dc,crOldText);
  }


void FXWDC::drawIconShaded(const FXIcon* icon,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::drawIconShaded: DC not connected to drawable.\n",getClassName()); }
  if(!icon || !icon->xid || !icon->shape){ fxerror("%s::drawIconShaded: illegal icon specified.\n",getClassName()); }
  if(!getApp()->grayBitmap || !getApp()->grayBitmap->xid){ fxerror("%s::drawIconShaded: standard gray bitmap not present.\n",getClassName()); }
  if(!xid){ fxerror("%s::drawIcon: DC not connected to drawable.\n",getClassName()); }
  if(!icon || !icon->xid || !icon->shape){ fxerror("%s::drawIcon: illegal icon specified.\n",getClassName()); }
  drawIcon(icon,dx,dy);
  }


void FXWDC::drawHashBox(FXint x,FXint y,FXint w,FXint h,FXint b){
  if(!xid){ fxerror("%s::drawHashBox: DC not connected to drawable.\n",getClassName()); }
  if(!getApp()->grayBitmap || !getApp()->grayBitmap->xid){ fxerror("%s::drawHashBox: standard gray bitmap not present.\n",getClassName()); }

  // Create the patterned brush
  HBRUSH hBrush=CreatePatternBrush((HBITMAP)getApp()->grayBitmap->xid);
  HBRUSH hOldBrush=(HBRUSH)SelectObject(dc,hBrush);
  HPEN hOldPen=(HPEN)SelectObject(dc,GetStockObject(NULL_PEN));

  // Draw the borders
  PatBlt(dc,x,y,w-b,b,PATINVERT);
  PatBlt(dc,x+w-b,y,b,h-b,PATINVERT);
  PatBlt(dc,x+b,y+h-b,w-b,b,PATINVERT);
  PatBlt(dc,x,y+b,b,h-b,PATINVERT);
  
  // Restore previous pen & brush
  SelectObject(dc,hOldPen);
  DeleteObject(SelectObject(dc,hOldBrush));
  }


void FXWDC::setForeground(FXColor clr){
  if(!xid){ fxerror("%s::setForeground: DC not connected to drawable.\n",getClassName()); }
  clr=colorpixel(clr);

  // Get current pen settings
  HPEN hCurPen=(HPEN)GetCurrentObject(dc,OBJ_PEN);
  EXTLOGPEN elp;
  GetObject(hCurPen,sizeof(EXTLOGPEN),&elp);

  // Create a new pen with same settings and new color
  LOGBRUSH lb;
  lb.lbStyle=BS_SOLID;
  lb.lbColor=clr;
  lb.lbHatch=0;
  DeleteObject(SelectObject(dc,ExtCreatePen(elp.elpPenStyle,elp.elpWidth,&lb,0,NULL)));

  // Reset the brush
  HBRUSH hBrush=(HBRUSH)SelectObject(dc,GetStockObject(NULL_BRUSH));
  GetObject(hBrush,sizeof(LOGBRUSH),&lb);
  DeleteObject(hBrush);
  lb.lbColor=clr;
  SelectObject(dc,CreateBrushIndirect(&lb));

  // And the text color
  SetTextColor(dc,clr);
  }


void FXWDC::setBackground(FXColor clr){
  if(!xid){ fxerror("%s::setBackground: DC not connected to drawable.\n",getClassName()); }
  clr=colorpixel(clr);
  SetBkColor(dc,clr);
  }


// Set dash pattern (for the LINE_ONOFF_DASH line style)
void FXWDC::setDashes(FXuint dashoffset,const FXchar *dashlist,FXuint n){
  if(!xid){ fxerror("%s::setDashes: DC not connected to drawable.\n",getClassName()); }
  }


void FXWDC::setLineWidth(FXuint linewidth){ 
  if(!xid){ fxerror("%s::setLineWidth: DC not connected to drawable.\n",getClassName()); }
  // Get current pen settings
  HPEN hCurPen=(HPEN)SelectObject(dc,GetStockObject(NULL_PEN));
  EXTLOGPEN elp;
  GetObject(hCurPen,sizeof(EXTLOGPEN),&elp);
  DeleteObject(hCurPen);

  // Create a new pen with this line width
  LOGBRUSH lb;
  lb.lbStyle=BS_SOLID;
  lb.lbColor=elp.elpColor;
  lb.lbHatch=0;
  SelectObject(dc,ExtCreatePen(elp.elpPenStyle,linewidth,&lb,0,NULL));
  }


void FXWDC::setLineCap(FXCapStyle capstyle){
  if(!xid){ fxerror("%s::setLineCap: DC not connected to drawable.\n",getClassName()); }
  // Get current pen settings
  HPEN hCurPen=(HPEN)SelectObject(dc,GetStockObject(NULL_PEN));
  EXTLOGPEN elp;
  GetObject(hCurPen,sizeof(EXTLOGPEN),&elp);
  DeleteObject(hCurPen);

  // Modify its cap style
  switch(capstyle){
    case CAP_NOT_LAST:
      dwPenType=PS_GEOMETRIC;
      dwEndCapAttributes=PS_ENDCAP_SQUARE;
      break;
    case CAP_BUTT:
      dwPenType=PS_GEOMETRIC;
      dwEndCapAttributes=PS_ENDCAP_FLAT;
      break;
    case CAP_ROUND:
      dwPenType=PS_GEOMETRIC;
      dwEndCapAttributes=PS_ENDCAP_ROUND;
      break;
    case CAP_PROJECTING:
      FXASSERT(FALSE);
      break;
    default:
      FXASSERT(FALSE);
    }

  // Create a new pen with this style
  LOGBRUSH lb;
  lb.lbStyle=BS_SOLID;
  lb.lbColor=elp.elpColor;
  lb.lbHatch=0;
  SelectObject(dc,ExtCreatePen(dwPenType|dwPenStyle|dwEndCapAttributes|dwJoinAttributes,elp.elpWidth,&lb,0,NULL));
  }


void FXWDC::setLineJoin(FXJoinStyle joinstyle){
  if(!xid){ fxerror("%s::setLineJoin: DC not connected to drawable.\n",getClassName()); }
  // Get current pen settings
  HPEN hCurPen=(HPEN)SelectObject(dc,GetStockObject(NULL_PEN));
  EXTLOGPEN elp;
  GetObject(hCurPen,sizeof(EXTLOGPEN),&elp);
  DeleteObject(hCurPen);

  // Now modify the style flags
  switch(joinstyle){
    case JOIN_MITER:
      dwPenType=PS_GEOMETRIC;
      dwJoinAttributes=PS_JOIN_MITER;
      break;
    case JOIN_ROUND:
      dwPenType=PS_GEOMETRIC;
      dwJoinAttributes=PS_JOIN_ROUND;
      break;
    case JOIN_BEVEL:
      dwPenType=PS_GEOMETRIC;
      dwJoinAttributes=PS_JOIN_BEVEL;
      break;
    default:
      FXASSERT(FALSE);
    }

  // Create a new pen with this style
  LOGBRUSH lb;
  lb.lbStyle=BS_SOLID;
  lb.lbColor=elp.elpColor;
  lb.lbHatch=0;
  SelectObject(dc,ExtCreatePen(dwPenType|dwPenStyle|dwEndCapAttributes|dwJoinAttributes,elp.elpWidth,&lb,0,NULL));
  }


void FXWDC::setLineStyle(FXLineStyle linestyle){
  if(!xid){ fxerror("%s::setLineStyle: DC not connected to drawable.\n",getClassName()); }
  // Get current pen settings
  HPEN hCurPen=(HPEN)SelectObject(dc,GetStockObject(NULL_PEN));
  EXTLOGPEN elp;
  GetObject(hCurPen,sizeof(EXTLOGPEN),&elp);
  DeleteObject(hCurPen);

  // Modify style bits
  switch(linestyle){
    case LINE_SOLID:
      dwPenType=PS_GEOMETRIC;
      dwPenStyle=PS_SOLID;
      break;
    case LINE_ONOFF_DASH:
      dwPenType=PS_COSMETIC;
      dwPenStyle=PS_DOT; // PS_ALTERNATE looks better, but is only supported for NT
      break;
    case LINE_DOUBLE_DASH:
      FXASSERT(FALSE);
      break;
    default:
      FXASSERT(FALSE);
    }

  // Create a new pen with this style
  LOGBRUSH lb;
  lb.lbStyle=BS_SOLID;
  lb.lbColor=elp.elpColor;
  lb.lbHatch=0;
  SelectObject(dc,ExtCreatePen(dwPenType|dwPenStyle|dwEndCapAttributes|dwJoinAttributes,elp.elpWidth,&lb,0,NULL));
  }


void FXWDC::setFillStyle(FXFillStyle fillstyle){
  if(!xid){ fxerror("%s::setFillStyle: DC not connected to drawable.\n",getClassName()); }
  // Get previous brush info
  HBRUSH hBrush=(HBRUSH) SelectObject(dc,GetStockObject(NULL_BRUSH));
  LOGBRUSH lb;
  GetObject(hBrush,sizeof(LOGBRUSH),&lb);
  DeleteObject(hBrush);

  // Set new style
  switch(fillstyle){
    case FILL_SOLID:
      lb.lbStyle=BS_SOLID;
      break;
    case FILL_TILED:
      FXASSERT(FALSE);
      break;
    case FILL_STIPPLED:
      FXASSERT(FALSE);
      break;
    case FILL_OPAQUESTIPPLED:
      lb.lbStyle=BS_PATTERN;
      lb.lbHatch=(LONG)stipplebitmap->id();
      break;
    default:
      FXASSERT(FALSE);
    }

  // Create the new brush
  SelectObject(dc,CreateBrushIndirect(&lb));
  }
  
  
// Set fill rule
void FXWDC::setFillRule(FXFillRule fillrule){
  if(!xid){ fxerror("%s::setFillRule: DC not connected to drawable.\n",getClassName()); }
  if(fillrule==RULE_EVEN_ODD)
    SetPolyFillMode(dc,ALTERNATE);
  else
    SetPolyFillMode(dc,WINDING);
  }


// Set blit function
void FXWDC::setFunction(FXFunction func){
  if(!xid){ fxerror("%s::setFunction: DC not connected to drawable.\n",getClassName()); }
  rop=func;
  }

// Set tile image
void FXWDC::setTile(FXImage* tile,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::setTile: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }

////////////// FIXME: likely not a good idea to hang on to the FXImage; the FOX object may go
//////////////        away even though the handle may hang around...

// Set stipple pattern
void FXWDC::setStipple(FXBitmap* stipple,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::setStipple: DC not connected to drawable.\n",getClassName()); }
  stipplebitmap=stipple;
  }

////////////// FIXME: likely not a good idea to hang on to the FXBitmap; the FOX object may go
//////////////        away even though the handle may hang around...

// Set clip rectangle
void FXWDC::setClipRectangle(FXint x,FXint y,FXint w,FXint h){
  if(!xid){ fxerror("%s::setClipRectangle: DC not connected to drawable.\n",getClassName()); }
  HRGN hrgn=CreateRectRgn(x,y,x+w,y+h);
  SelectClipRgn(dc,hrgn);
  }
 

// Clear clipping
void FXWDC::clearClipRectangle(){
  if(!xid){ fxerror("%s::clearClipRectangle: DC not connected to drawable.\n",getClassName()); }
  SelectClipRgn(dc,NULL);
  }


// Set clip mask
void FXWDC::setClipMask(FXBitmap* mask,FXint dx,FXint dy){
  if(!xid){ fxerror("%s::setClipMask: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }


// Clear clip mask
void FXWDC::clearClipMask(){
  if(!xid){ fxerror("%s::clearClipMask: DC not connected to drawable.\n",getClassName()); }
  FXASSERT(FALSE);
  }

  
// Clip drawing by subwindows
void FXWDC::clipChildWindows(FXbool yes){ 
  if(!xid){ fxerror("%s::clipChildWindows: DC not connected to drawable.\n",getClassName()); }
  DWORD dwStyle=GetWindowLong((HWND)xid,GWL_STYLE);
  if(yes){
    dwStyle|=WS_CLIPCHILDREN;
    }
  else{
    dwStyle&=~WS_CLIPCHILDREN;
    }
  SetWindowLong((HWND)xid,GWL_STYLE,dwStyle);
  }


// Set font to draw text with
void FXWDC::setTextFont(FXFont *font){
  if(!xid){ fxerror("%s::setTextFont: DC not connected to drawable.\n",getClassName()); }
  if(!font || !font->id()){ fxerror("%s::setTextFont: illegal or NULL font specified.\n",getClassName()); }
  SelectObject(dc,font->id());
  }


#endif
