/********************************************************************************
*                                                                               *
*                            W i n d o w   O b j e c t                          *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997 by Jeroen van der Zijp.   All Rights Reserved.             *
*********************************************************************************
* Major Contributions for Windows NT by Lyle Johnson                            *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
*********************************************************************************
* $Id: FXWindow.cpp,v 1.159 1999/01/15 04:54:14 jeroen Exp $                    *
********************************************************************************/
#include "xincs.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXDict.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXObjectList.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXWDC.h"
#include "FXCursor.h"
#include "FXDrawable.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXRootWindow.h"
#include "FXShell.h"

/*
 To Do:
  - Forward various messages to target
  - Look at destroy(); cannot call overload from base class's destructor!
  - Add virtual overloads for colors (as colors depend on map!)
  - When window is disabled, it should loose focus
  - When no subclass handles SEL_SELECTION_REQUEST, send back None property here
  - The jury on a special FXColormap class is still out.
  - Add URL jump text also.
  - Update should only happen if widget is not in some sort of transaction.
  - Not every widget needs help and tip data; move this down to buttons etc.
  - Should sole modifier keys be ignored before attempting handling through acceltable?
  - Need clipboard support too.
  - Should we catch & translate Escape key & turn it into a SEL_CANCEL message?
  - Should window have a pointer to the main (shell-) window, e.g. for popups, transientfors...?
  - Default constructors [for serialization] should:
      - Initialize dynamic member variables same as regular constructor.
        Dynamic variables are those that are not saved during serialization.
      - Initialize non-dynamic variables with ``garbage.''  Those variables
        are saved during serialization, and should be loaded back also.
  - We initialize/deinitialize with ``garbage'' to bring latent software bugs
    in the application codes to light as early as possible; this is a permanent
    feature in FOX, not just for compiles in debug mode.
  - If FLAG_DIRTY gets reset at the END of layout(), it will be safe to have
    show() and hide() to call recalc(), in case they get called from the
    application code.
  - Need some sort of journal facility.
  - Use INCR mechanism if it gets large.
  - Consolidate FXWindow and FXApp DND stuff into one place, if possible.
  - Change: passModalEvents() should not rely on FXPopups only; perhaps a flag?
  - When someone requests a selection with target ddeDelete, the selection should
    be deleted!
  - Perhaps FXWindow, not FXApp, should have a pointer to FXDC.
  - Resizing from non-0 to 0 should hide, and from 0 to non-0 show; thus sizes of 0 may 
    become legal!
*/

#ifndef FX_NATIVE_WIN32

// Basic events
#define BASIC_EVENT_MASK   (ExposureMask|StructureNotifyMask|PropertyChangeMask|EnterWindowMask|LeaveWindowMask|VisibilityChangeMask|KeyPressMask|KeyReleaseMask)

// Additional events for shell widget events
#define SHELL_EVENT_MASK   (FocusChangeMask|StructureNotifyMask)

// Additional events for enabled widgets        
#define ENABLED_EVENT_MASK (ButtonPressMask|ButtonReleaseMask|PointerMotionMask)

// These events are grabbed for mouse grabs
#define GRAB_EVENT_MASK    (ButtonPressMask|ButtonReleaseMask|PointerMotionMask|EnterWindowMask|LeaveWindowMask|VisibilityChangeMask)

// Do not propagate mask
#define NOT_PROPAGATE_MASK (KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask|ButtonMotionMask)

#endif


/*******************************************************************************/

// Map
FXDEFMAP(FXWindow) FXWindowMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXWindow::onPaint),
  FXMAPFUNC(SEL_UPDATE,0,FXWindow::onUpdate),
  FXMAPFUNC(SEL_ENTER,0,FXWindow::onEnter),
  FXMAPFUNC(SEL_LEAVE,0,FXWindow::onLeave),
  FXMAPFUNC(SEL_DESTROY,0,FXWindow::onDestroy),
  FXMAPFUNC(SEL_VISIBILITY,0,FXWindow::onVisibility),
  FXMAPFUNC(SEL_CONFIGURE,0,FXWindow::onConfigure),
  FXMAPFUNC(SEL_MAP,0,FXWindow::onMap),
  FXMAPFUNC(SEL_UNMAP,0,FXWindow::onUnmap),
  FXMAPFUNC(SEL_FOCUSIN,0,FXWindow::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXWindow::onFocusOut),
  FXMAPFUNC(SEL_SELECTION_LOST,0,FXWindow::onSelectionLost),
  FXMAPFUNC(SEL_SELECTION_GAINED,0,FXWindow::onSelectionGained),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,FXWindow::onSelectionRequest),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXWindow::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXWindow::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXWindow::onMiddleBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXWindow::onMiddleBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXWindow::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXWindow::onRightBtnRelease),
  FXMAPFUNC(SEL_UNGRABBED,0,FXWindow::onUngrabbed),
  FXMAPFUNC(SEL_KEYPRESS,0,FXWindow::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXWindow::onKeyRelease),
  FXMAPFUNC(SEL_KEYPRESS,FXWindow::ID_HOTKEY,FXWindow::onHotKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,FXWindow::ID_HOTKEY,FXWindow::onHotKeyRelease),
  FXMAPFUNC(SEL_DND_ENTER,0,FXWindow::onDNDEnter),
  FXMAPFUNC(SEL_DND_LEAVE,0,FXWindow::onDNDLeave),
  FXMAPFUNC(SEL_DND_DROP,0,FXWindow::onDNDDrop),
  FXMAPFUNC(SEL_DND_MOTION,0,FXWindow::onDNDMotion),
  FXMAPFUNC(SEL_FOCUS_SELF,0,FXWindow::onFocusSelf),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_TOGGLESHOWN,FXWindow::onUpdToggleShown),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_BACK_COLOR,FXWindow::onUpdBackColor),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_DELETE,FXWindow::onUpdYes),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SHOW,FXWindow::onCmdShow),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_HIDE,FXWindow::onCmdHide),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_TOGGLESHOWN,FXWindow::onCmdToggleShown),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_RAISE,FXWindow::onCmdRaise),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_LOWER,FXWindow::onCmdLower),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_ENABLE,FXWindow::onCmdEnable),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_DISABLE,FXWindow::onCmdDisable),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_UPDATE,FXWindow::onCmdUpdate),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_BACK_COLOR,FXWindow::onCmdBackColor),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_DELETE,FXWindow::onCmdDelete),
  };


// Object implementation
FXIMPLEMENT(FXWindow,FXDrawable,FXWindowMap,ARRAYNUMBER(FXWindowMap))
  
  
/*******************************************************************************/

  
// Request for the widgets supported DND types
const FXchar FXWindow::targetTypeName[]="TARGETS";


// Request the widgets data will be deleted
const FXchar FXWindow::deleteTypeName[]="DELETE";


// Request the widgets ascii text value
const FXchar FXWindow::textTypeName[]="text/plain";

// Name of the drag type
const FXchar FXWindow::colorTypeName[]="application/x-color";


// Atom value for type list
FXDragType FXWindow::targetType=0;


// Atom value for delete 
FXDragType FXWindow::deleteType=0;


// Atom value for ascii text type
FXDragType FXWindow::textType=0;


// Color drag type
FXDragType FXWindow::colorType=0;


/*******************************************************************************/

  
// For deserialization
FXWindow::FXWindow(){
  parent=(FXWindow*)-1;
  shell=(FXWindow*)-1;
  first=(FXWindow*)-1;
  last=(FXWindow*)-1;
  next=(FXWindow*)-1;
  prev=(FXWindow*)-1;
  focus=NULL;
  defaultCursor=(FXCursor*)-1;
  dragCursor=(FXCursor*)-1;
  accelTable=(FXAccelTable*)-1;
  target=NULL;
  message=0;
  xpos=0;
  ypos=0;
  backColor=getApp()->backColor;
  flags=FLAG_DIRTY|FLAG_UPDATE|FLAG_RECALC|FLAG_CLIPCHILDREN;
  options=0;
  }


// Only used for the Root Window
FXWindow::FXWindow(FXApp* a):FXDrawable(a,1,1,0){
  parent=NULL;
  shell=this;
  first=last=NULL;
  next=prev=NULL;
  focus=NULL;
  target=NULL;
  message=0;
  defaultCursor=getApp()->arrowCursor;
  dragCursor=getApp()->arrowCursor;
  accelTable=NULL;
  xpos=0;
  ypos=0;
  backColor=getApp()->backColor;
  flags=FLAG_DIRTY|FLAG_SHOWN|FLAG_MAPPED|FLAG_UPDATE|FLAG_RECALC|FLAG_CLIPCHILDREN;
  options=LAYOUT_FIX_X|LAYOUT_FIX_Y|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT;
  }


// This constructor is used for Shell Windows
FXWindow::FXWindow(FXApp* a,FXuint opts,FXint x,FXint y,FXint w,FXint h):FXDrawable(a,w,h,0){
  parent=a->root;
  shell=this;
  first=last=NULL;
  prev=parent->last;
  next=NULL;
  parent->last=this;
  if(prev) prev->next=this; else parent->first=this;
  focus=NULL;
  target=NULL;
  message=0;
  defaultCursor=getApp()->arrowCursor;
  dragCursor=getApp()->arrowCursor;
  accelTable=NULL;
  xpos=x;
  ypos=y;
  backColor=getApp()->backColor;
  flags=FLAG_DIRTY|FLAG_UPDATE|FLAG_RECALC|FLAG_CLIPCHILDREN;
  options=opts;
  }


// This constructor is used for all child-windows
FXWindow::FXWindow(FXComposite* p,FXuint opts,FXint x,FXint y,FXint w,FXint h):FXDrawable(p->getApp(),w,h,0){
  parent=p;
  shell=p->getShell();
  first=last=NULL;
  prev=parent->last;
  next=NULL;
  parent->last=this;
  if(prev) prev->next=this; else parent->first=this;
  focus=NULL;
  target=NULL;
  message=0;
  defaultCursor=getApp()->arrowCursor;
  dragCursor=getApp()->arrowCursor;
  accelTable=NULL;
  xpos=x;
  ypos=y;
  backColor=getApp()->backColor;
  flags=FLAG_DIRTY|FLAG_UPDATE|FLAG_RECALC|FLAG_CLIPCHILDREN;
  options=opts;
  }


// Return DC for this surface
FXDC* FXWindow::DC() const {
  return getApp()->windowdc;
  }


/*******************************************************************************/


// Save data
void FXWindow::save(FXStream& store) const {
  FXDrawable::save(store);
  store << parent;
  store << shell;
  store << first;
  store << last;
  store << next;
  store << prev;
  store << focus;
  store << defaultCursor;
  store << dragCursor;
  store << accelTable;
  store << target;        // How to resolve?
  store << message;
  store << xpos;
  store << ypos;
  store << backColor;
  store << options;
  }


// Load data
void FXWindow::load(FXStream& store){ 
  FXDrawable::load(store);
  store >> parent;
  store >> shell;
  store >> first;
  store >> last;
  store >> next;
  store >> prev;
  store >> focus;
  store >> defaultCursor;
  store >> dragCursor;
  store >> accelTable;
  store >> target;        // How to resolve?
  store >> message;
  store >> xpos;
  store >> ypos;
  store >> backColor;
  store >> options;
  }


/*******************************************************************************/


// Handle repaint 
long FXWindow::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXDC *dc=DC();
  dc->begin(this);
  dc->setForeground(backColor);
  dc->fillRectangle(ev->rect.x,ev->rect.y,ev->rect.w,ev->rect.h);
  dc->end();
  return 1;
  }


// Keep track of visibility changes
long FXWindow::onVisibility(FXObject*,FXSelector,void* ptr){
  FXEvent *ev=(FXEvent*)ptr;
  FXTRACE((250,"%s::onVisibility %08x %s\n",getClassName(),this,(ev->code==VISIBILITYTOTAL)?"total":(ev->code==VISIBILITYPARTIAL)?"partial":"none"));
  flags&=~(FLAG_OBSCURED|FLAG_INVISIBLE);
  if(ev->code==VISIBILITYPARTIAL) flags|=FLAG_OBSCURED;
  if(ev->code==VISIBILITYNONE) flags|=(FLAG_OBSCURED|FLAG_INVISIBLE);
  return 1;
  }


// Keep track of mapped state
long FXWindow::onMap(FXObject*,FXSelector,void*){
  FXTRACE((250,"%s::onMap %08x\n",getClassName(),this));
  flags|=FLAG_MAPPED;
  return 0;
  }


// Keep track of mapped state
long FXWindow::onUnmap(FXObject*,FXSelector,void*){
  FXTRACE((250,"%s::onUnmap %08x\n",getClassName(),this));
  if(getApp()->grabWindow==this) getApp()->grabWindow=NULL;
  flags&=~FLAG_MAPPED;
  return 0;
  }


// Handle configure notify
long FXWindow::onConfigure(FXObject*,FXSelector,void*){
  FXTRACE((250,"%s::onConfigure %08x\n",getClassName(),this));
  return 0;
  }


// Handle destroy event
long FXWindow::onDestroy(FXObject*,FXSelector,void*){
  FXTRACE((250,"%s::onDestroy %08x\n",getClassName(),this));
#ifndef FX_NATIVE_WIN32
  XDeleteContext(getDisplay(),xid,getApp()->wcontext);
#endif
  if(getApp()->grabWindow==this) getApp()->grabWindow=NULL;
  flags&=~(FLAG_MAPPED|FLAG_SHOWN|FLAG_FOCUSED);
  xid=0;
  return 1;
  }


// Hot key combination pressed
long FXWindow::onHotKeyPress(FXObject*,FXSelector,void* ptr){
  FXTRACE((200,"%s::onHotKeyPress %08x\n",getClassName(),this));
  flags&=~FLAG_TIP;     // Kill tip
  if(isEnabled()){
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    return handle(this,MKUINT(0,SEL_ACTIVATE),ptr);
    }
  return 0;
  }


// Hot key combination released
long FXWindow::onHotKeyRelease(FXObject*,FXSelector,void* ptr){
  FXTRACE((200,"%s::onHotKeyRelease %08x\n",getClassName(),this));
  flags&=~FLAG_TIP;
  if(isEnabled()){
    return handle(this,MKUINT(0,SEL_DEACTIVATE),ptr);
    }
  return 0;
  }


// Handle incoming button events; generally, we give the target a 
// first shot at it; if it doesn't handle it, we'll bounce it around
// as an activate/deactivate event to this widget, after we
// set the focus to this widget!
long FXWindow::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    return handle(this,MKUINT(0,SEL_ACTIVATE),ptr);
    }
  return 0;
  }

long FXWindow::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    return handle(this,MKUINT(0,SEL_DEACTIVATE),ptr);
    }
  return 0;
  }

long FXWindow::onMiddleBtnPress(FXObject*,FXSelector,void* ptr){
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    return 1;
    }
  return 0;
  }

long FXWindow::onMiddleBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONRELEASE),ptr)) return 1;
    return 1;
    }
  return 0;
  }

long FXWindow::onRightBtnPress(FXObject*,FXSelector,void* ptr){
  flags&=~FLAG_TIP;
  if(isEnabled()){
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONPRESS),ptr)) return 1;
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    return 1;
    }
  return 0;
  }

long FXWindow::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    return handle(this,MKUINT(ID_QUERY_MENU,SEL_COMMAND),ptr);
    }
  return 0;
  }


// Keyboard press; forward to focus child
long FXWindow::onKeyPress(FXObject*,FXSelector sel,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  FXTRACE((200,"%s::onKeyPress keysym=0x%04x state=%04x\n",getClassName(),event->code,event->state));
  
  // Kill the tip
  flags&=~FLAG_TIP;
  
  // Try bounce to the target first
  if(!getFocus() && isEnabled() && target && target->handle(this,MKUINT(message,SEL_KEYPRESS),ptr)) return 1;
  
  // Next, try current focus widget
  if(getFocus() && getFocus()->handle(focus,sel,ptr)) return 1;
  
// #ifndef FX_NATIVE_WIN32
//   XAutoRepeatOff(getDisplay());
// #endif
  
  // Next, check the accelerators...
  if(getAccelTable() && getAccelTable()->handle(this,sel,ptr)) return 1;
  
  // Otherwise, perform the routine keyboard processing; ignore modifiers
  switch(event->code){
    case KEY_Shift_L:
    case KEY_Shift_R:
    case KEY_Control_L:
    case KEY_Control_R:
    case KEY_Caps_Lock:
    case KEY_Shift_Lock:
    case KEY_Meta_L:
    case KEY_Meta_R:
    case KEY_Alt_L:
    case KEY_Alt_R:
    case KEY_Super_L:
    case KEY_Super_R:
    case KEY_Hyper_L:
    case KEY_Hyper_R:
    case KEY_Scroll_Lock:
    case KEY_Sys_Req:
      return 0;
    case KEY_space:
    case KEY_KP_Enter:
    case KEY_Return:
      if(isEnabled() && handle(this,MKUINT(0,SEL_ACTIVATE),ptr)) return 1;
      break;
    case KEY_Tab:
      if(event->state&SHIFTMASK) goto prv;
    case KEY_Next:
      return handle(this,MKUINT(0,SEL_FOCUS_NEXT),ptr);
    case KEY_Prior: 
    case KEY_ISO_Left_Tab:
prv:  return handle(this,MKUINT(0,SEL_FOCUS_PREV),ptr);
    case KEY_Up:    
      return handle(this,MKUINT(0,SEL_FOCUS_UP),ptr);
    case KEY_Down:  
      return handle(this,MKUINT(0,SEL_FOCUS_DOWN),ptr);
    case KEY_Left:  
      return handle(this,MKUINT(0,SEL_FOCUS_LEFT),ptr);
    case KEY_Right: 
      return handle(this,MKUINT(0,SEL_FOCUS_RIGHT),ptr);
    case KEY_Home:
      return handle(this,MKUINT(0,SEL_FOCUS_HOME),ptr);
    case KEY_End:
      return handle(this,MKUINT(0,SEL_FOCUS_END),ptr);
    }
  return 0;
  }


// Keyboard release; sent to focus widget
long FXWindow::onKeyRelease(FXObject*,FXSelector sel,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  FXTRACE((200,"%s::onKeyRelease keysym=0x%04x\n",getClassName(),event->code));
  
  // Try bounce to the target first
  if(!getFocus() && isEnabled() && target && target->handle(this,MKUINT(message,SEL_KEYRELEASE),ptr)) return 1;

  // Next bounce to focus widget
  if(getFocus() && getFocus()->handle(focus,sel,ptr)) return 1;
  
// #ifndef FX_NATIVE_WIN32
//   XAutoRepeatOn(getDisplay());
// #endif
  
  // Next, check the accelerators...
  if(getAccelTable() && getAccelTable()->handle(this,sel,ptr)) return 1;

  // Otherwise, perform the routine keyboard processing
  switch(event->code){
    case KEY_Shift_L:
    case KEY_Shift_R:
    case KEY_Control_L:
    case KEY_Control_R:
    case KEY_Caps_Lock:
    case KEY_Shift_Lock:
    case KEY_Meta_L:
    case KEY_Meta_R:
    case KEY_Alt_L:
    case KEY_Alt_R:
    case KEY_Super_L:
    case KEY_Super_R:
    case KEY_Hyper_L:
    case KEY_Hyper_R:
    case KEY_Scroll_Lock:
    case KEY_Sys_Req:
      return 0;
    case KEY_space:
    case KEY_KP_Enter:
    case KEY_Return:
      if(isEnabled() && handle(this,MKUINT(0,SEL_DEACTIVATE),ptr)) return 1;
      break;
    }
  return 0;
  }


// Entering window
long FXWindow::onEnter(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  FXTRACE((250,"%s::onEnter %08x\n",getClassName(),this));
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_ENTER),ptr)) return 1;
    }
  getApp()->cursorWindow=this;
  flags|=(FLAG_CURSOR|FLAG_HELP);
  if(event->state==0) flags|=FLAG_TIP;
  return 1;
  }


// Leaving window
long FXWindow::onLeave(FXObject*,FXSelector,void* ptr){
  FXTRACE((250,"%s::onLeave %08x\n",getClassName(),this));
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_LEAVE),ptr)) return 1;
    }
  getApp()->cursorWindow=parent;
  flags&=~(FLAG_CURSOR|FLAG_TIP|FLAG_HELP);
  return 1;
  }


// True if window under the cursor
FXbool FXWindow::underCursor() const {
  return (flags&FLAG_CURSOR)!=0;
  }


// Gained focus
long FXWindow::onFocusIn(FXObject*,FXSelector,void* ptr){
  FXTRACE((250,"%s::onFocusIn %08x\n",getClassName(),this));
  if(focus) focus->handle(focus,MKUINT(0,SEL_FOCUSIN),NULL);
  if(target) target->handle(this,MKUINT(message,SEL_FOCUSIN),ptr);
  flags|=FLAG_FOCUSED;
  return 1;
  }


// Lost focus
long FXWindow::onFocusOut(FXObject*,FXSelector,void* ptr){
  FXTRACE((250,"%s::onFocusOut %08x\n",getClassName(),this));
  if(focus) focus->handle(focus,MKUINT(0,SEL_FOCUSOUT),NULL);
  if(target) target->handle(this,MKUINT(message,SEL_FOCUSOUT),ptr);
  flags&=~FLAG_FOCUSED;
  return 1;
  }


// Handle drag-and-drop enter
long FXWindow::onDNDEnter(FXObject*,FXSelector,void* ptr){
  FXTRACE((100,"%s::onDNDEnter %08x\n",getClassName(),this));
  if(isDropEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_DND_ENTER),ptr)) return 1;
    }
  return 0;
  }


// Handle drag-and-drop leave
long FXWindow::onDNDLeave(FXObject*,FXSelector,void* ptr){
  FXTRACE((100,"%s::onDNDLeave %08x\n",getClassName(),this));
  if(isDropEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_DND_LEAVE),ptr)) return 1;
    }
  return 0;
  }


// Handle drag-and-drop motion
long FXWindow::onDNDMotion(FXObject*,FXSelector,void* ptr){
  FXTRACE((100,"%s::onDNDMotion %08x\n",getClassName(),this));
  if(isDropEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_DND_MOTION),ptr)) return 1;
    }
  return 0;
  }


// Handle drag-and-drop drop
long FXWindow::onDNDDrop(FXObject*,FXSelector,void* ptr){
  FXTRACE((100,"%s::onDNDDrop %08x\n",getClassName(),this));
  if(isDropEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_DND_DROP),ptr)) return 1;
    }
  return 0;
  }


// Focus on widget itself
long FXWindow::onFocusSelf(FXObject*,FXSelector,void* ptr){
  FXTRACE((250,"%s::onFocusSelf %08x\n",getClassName(),this));
  setFocus();
  return 1;
  }


// Show window
long FXWindow::onCmdShow(FXObject*,FXSelector,void*){ 
  if(!shown()){ show(); recalc(); }
  return 1; 
  }


// Hide window
long FXWindow::onCmdHide(FXObject*,FXSelector,void*){ 
  if(shown()){ hide(); recalc(); }
  return 1; 
  }

// Hide or show window
long FXWindow::onCmdToggleShown(FXObject*,FXSelector,void*){ 
  shown() ? hide() : show(); 
  recalc(); 
  return 1;
  }


// Update hide or show window 
long FXWindow::onUpdToggleShown(FXObject* sender,FXSelector,void*){ 
  FXuint msg=shown() ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }


// Raise window
long FXWindow::onCmdRaise(FXObject*,FXSelector,void*){ 
  raise(); 
  return 1; 
  }


// Lower window
long FXWindow::onCmdLower(FXObject*,FXSelector,void*){ 
  lower(); 
  return 1; 
  }


// Delete window
long FXWindow::onCmdDelete(FXObject*,FXSelector,void*){ 
  delete this; 
  return 1; 
  }


// Enable window
long FXWindow::onCmdEnable(FXObject*,FXSelector,void*){ 
  enable(); 
  return 1; 
  }


// Disable window
long FXWindow::onCmdDisable(FXObject*,FXSelector,void*){ 
  disable(); 
  return 1; 
  }


// In combination with autograying/autohiding widgets,
// this could be used to ungray or show when you don't want
// to write a whole handler routine just to do this.
long FXWindow::onUpdYes(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  return 1;
  }


// Update (repaint) window
long FXWindow::onCmdUpdate(FXObject*,FXSelector,void*){ 
  update(0,0,width,height); 
  return 1; 
  }


// Set back color
long FXWindow::onCmdBackColor(FXObject*,FXSelector,void* ptr){
  setBackColor((FXColor)ptr);
  return 1;
  }


// Set back color
long FXWindow::onUpdBackColor(FXObject* sender,FXSelector,void* ptr){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)backColor);
  return 1;
  }


/*******************************************************************************/

// If window can have focus
FXbool FXWindow::canFocus() const { 
  return FALSE; 
  }


// Has window the focus
FXbool FXWindow::hasFocus() const { 
  return (flags&FLAG_FOCUSED)!=0; 
  }


// Set focus to this widget.
// The chain of focus from shell down to a control is changed.
// Widgets now in the chain may or may not gain real focus,
// depending on whether parent window already had a real focus!
// Setting the focus to a composite will cause descendants to loose it.
void FXWindow::setFocus(){
  if(shell==this) return;
  if(parent->focus==this){
    if(focus) focus->killFocus();
    }
  else{
    if(parent->focus) 
      parent->focus->killFocus();
    else
      parent->setFocus();
    parent->focus=this;
    if(parent->hasFocus()) handle(this,MKUINT(0,SEL_FOCUSIN),NULL);
    }
  getApp()->focusWindow=this;
  flags|=FLAG_HELP;
  }


// Kill focus to this widget.
void FXWindow::killFocus(){
  if(shell==this) return;
  if(parent->focus!=this) return;
  if(focus) focus->killFocus();
  if(hasFocus()) handle(this,MKUINT(0,SEL_FOCUSOUT),NULL);
  parent->focus=NULL;
  getApp()->focusWindow=NULL;
  flags&=~FLAG_HELP;
  }


/*******************************************************************************/


// Create X window
void FXWindow::create(){ 
  if(!xid){
    FXTRACE((150,"%s::create %08x\n",getClassName(),this));
    
#ifndef FX_NATIVE_WIN32
    
    XSetWindowAttributes wattr;
    unsigned long mask;
    Visual* visual;
    
    // Gotta have display open!
    if(!getDisplay()){ fxerror("%s::create: trying to create window before opening display.\n",getClassName()); }
  
    // Gotta have a parent already created!
    if(!parent->id()){ fxerror("%s::create: trying to create window before creating parent window.\n",getClassName()); }
      
    // Create default cursor
    if(defaultCursor) defaultCursor->create();
    
    // Create drag cursor
    if(dragCursor) dragCursor->create();
  
    // Fill in the attributes
    mask=CWBackPixmap|CWWinGravity|CWBitGravity|CWBorderPixel|CWEventMask|CWDontPropagate|CWCursor|CWOverrideRedirect|CWSaveUnder|CWColormap;
    
    // Events for normal windows
    wattr.event_mask=BASIC_EVENT_MASK;
    
    // Events for shell windows
    if(shell==this) wattr.event_mask|=SHELL_EVENT_MASK;
    
    // If enabled, turn on some more events
    if(flags&FLAG_ENABLED) wattr.event_mask|=ENABLED_EVENT_MASK;
    
    // FOX will not propagate events to ancestor windows
    wattr.do_not_propagate_mask=NOT_PROPAGATE_MASK;

    // Determine visual first
    visual=getDefaultVisual();
    
    // Next, determine depth
    depth=getDefaultDepth();
    
    // Next, determine colormap
    wattr.colormap=getDefaultColormap();
    
    // This is needed for OpenGL
    wattr.border_pixel=0;
    
    // No background
    wattr.background_pixmap=None;
    
    // We don't seem to be able to do ForgetGravity for win_gravity
    // Its the same as UnmapGravity, which is useless...
    wattr.bit_gravity=ForgetGravity;
    wattr.win_gravity=NorthWestGravity;
    
    // Determine override redirect
    wattr.override_redirect=doesOverrideRedirect();
    
    // Determine save-unders
    wattr.save_under=doesSaveUnder();
  
    // Set cursor  
    wattr.cursor=defaultCursor->id();
    
    // Finally, create the window
    xid=XCreateWindow(getDisplay(),parent->id(),xpos,ypos,FXMAX(width,1),FXMAX(height,1),0,depth,InputOutput,visual,mask,&wattr);
    
    // Uh-oh, we failed
    if(!xid){ fxerror("%s::create: unable to create window.\n",getClassName()); }
  
    // For dialogs etc, set the transient for
    ///XSetTransientForHint(getDisplay(),xid,?); ////////// FIX

    // Store for xid to C++ object mapping
    XSaveContext(getDisplay(),xid,getApp()->wcontext,(XPointer)this);
    
#else
    
    // Gotta have a parent already created!
    if(!parent->id()){ fxerror("%s::create: trying to create window before creating parent window.\n",getClassName()); }

    // Next, determine depth
    depth=getDefaultDepth();

    // Finally, create the window
    DWORD dwStyle=WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN;
    HWND hParent=(HWND)parent->id();
    if(shell==this){

      // Make sure popups are owned by us, not by the desktop;
      // Otherwise they appear as separate buttons on the taskbar.
      if(hParent==GetDesktopWindow()){
	if(getApp()->getMainWindow()->id()) hParent=(HWND)getApp()->getMainWindow()->id();
	}

      if(doesOverrideRedirect()){
        dwStyle=WS_POPUP;
	}
      else{
	dwStyle=WS_OVERLAPPEDWINDOW;
	}
      }

    xid=CreateWindow("FXWindow",NULL,dwStyle,xpos,ypos,FXMAX(width,1),FXMAX(height,1),hParent,NULL,FXApp::hInstance,this);

    // Uh-oh, we failed
    if(!xid){ fxerror("%s::create: unable to create window.\n",getClassName()); }

#endif
    }
  }


// Destroy
void FXWindow::destroy(){
  if(xid){
    
    FXTRACE((150,"%s::destroy %08x\n",getClassName(),this));
    
#ifndef FX_NATIVE_WIN32

    // Delete the window
    XDestroyWindow(getDisplay(),xid);
    
    // Remove from xid to C++ object mapping
    XDeleteContext(getDisplay(),xid,getApp()->wcontext);

#else
    
    // Zap the window
    DestroyWindow((HWND)xid);
    
#endif
    
    // No longer grabbed
    if(getApp()->grabWindow==this) getApp()->grabWindow=NULL;
    
    xid=0;
    }
  }


/*******************************************************************************/

// Get default width
FXint FXWindow::getDefaultWidth(){ 
  return 1; 
  }


// Get default height
FXint FXWindow::getDefaultHeight(){ 
  return 1; 
  }


// Window is the default target
FXbool FXWindow::isDefault() const { 
  return (flags&FLAG_DEFAULT)!=0; 
  }


// Test if active
FXbool FXWindow::isActive() const {
  return (flags&FLAG_ACTIVE)!=0;
  }


// Set X position
void FXWindow::setX(FXint x){ 
  xpos=x; 
  recalc();
  }


// Set Y position
void FXWindow::setY(FXint y){ 
  ypos=y; 
  recalc();
  }


// Set width
void FXWindow::setWidth(FXint w){ 
  width=FXMAX(w,1); 
  recalc();
  }


// Set height
void FXWindow::setHeight(FXint h){ 
  height=FXMAX(h,1); 
  recalc();
  }


// Change layout
void FXWindow::setLayoutHints(FXuint lout){
  FXuint opts=(options&~LAYOUT_MASK) | (lout&LAYOUT_MASK);
  if(options!=opts){
    options=opts;
    recalc();
    }
  }


// Get layout hints
FXuint FXWindow::getLayoutHints() const { 
  return (options&LAYOUT_MASK); 
  }


// Is widget a composite
FXbool FXWindow::isComposite() const { 
  return 0; 
  }


// Return visual
#ifndef FX_NATIVE_WIN32
Visual* FXWindow::getDefaultVisual(){ 
  return DefaultVisual(getDisplay(),DefaultScreen(getDisplay())); 
  }
#endif


// Get depth
FXint FXWindow::getDefaultDepth(){
#ifndef FX_NATIVE_WIN32
  return DefaultDepth(getDisplay(),DefaultScreen(getDisplay()));
#else
  HDC hdc=GetDC((HWND)xid);
  FXint nDepth=GetDeviceCaps(hdc,BITSPIXEL);
  ReleaseDC((HWND)xid,hdc);
  return nDepth;
#endif
  }


// Return colormap
#ifndef FX_NATIVE_WIN32
FXID FXWindow::getDefaultColormap(){ 
  return DefaultColormap(getDisplay(),DefaultScreen(getDisplay())); 
  }
#endif


// Window does override-redirect
FXbool FXWindow::doesOverrideRedirect() const { 
  return FALSE; 
  }


// Window does save-unders
FXbool FXWindow::doesSaveUnder() const { 
  return FALSE; 
  }


// Modal events to this window should be passed
FXbool FXWindow::passModalEvents() const {
  // We're assuming [for now] only popups do save-unders!!!
  return !getApp()->invocation || getApp()->invocation->window==shell->xid || shell->doesSaveUnder();
  }


// Add hot key to closest ancestor's accelerator table
void FXWindow::addHotKey(FXHotKey code){
  FXAccelTable *accel; FXWindow *win=this;
  while((accel=win->getAccelTable())==NULL && win!=shell) win=win->parent;
  if(accel) accel->addAccel(code,this,MKUINT(ID_HOTKEY,SEL_KEYPRESS),MKUINT(ID_HOTKEY,SEL_KEYRELEASE));
  }


// Remove hot key from closest ancestor's accelerator table
void FXWindow::remHotKey(FXHotKey code){
  FXAccelTable *accel; FXWindow *win=this;
  while((accel=win->getAccelTable())==NULL && win!=shell) win=win->parent;
  if(accel) accel->removeAccel(code);
  }


// Check if child is related
FXbool FXWindow::containsChild(const FXWindow* child) const {
  while(child){ 
    if(child==this) return 1;
    child=child->parent;
    }
  return 0;
  }


// Check if window is related to parent window
FXbool FXWindow::isChildOf(const FXWindow* window) const {
  const FXWindow* w=this;
  while(w){
    if(w==window) return 1;
    w=w->parent;
    }
  return 0;
  }


// Get child at x,y
FXWindow* FXWindow::getChildAt(FXint x,FXint y) const {
  register FXWindow* child;
  for(child=first; child; child=child->next){
    if(child->shown() && child->FXWindow::contains(x,y)) return child;
    }
  return NULL;
  }


// Count number of children
FXint FXWindow::numChildren() const {
  register const FXWindow *child=first;
  register FXint num=0;
  while(child){
    child=child->next;
    num++;
    }
  return num;
  }


// Get index of child window
FXint FXWindow::indexOfChild(const FXWindow *window) const {
  register FXint index=0;
  if(!window) return -1;
  while(window->prev){
    window=window->prev;
    index++;
    }
  return index;
  }


// Get child window at index
FXWindow* FXWindow::childAtIndex(FXint index) const {
  register FXWindow* child=first;
  if(index<0) return NULL;
  while(index && child){
    child=child->next;
    index--;
    }
  return child;
  }


/*******************************************************************************/

#ifndef FX_NATIVE_WIN32

// Add this window to the list of colormap windows
void FXWindow::addColormapWindows(){
  Window windows[2],*windowsReturn,*windowList;
  int countReturn;
  
  // Check to see if there is already a property
  Status status=XGetWMColormapWindows(getDisplay(),shell->id(),&windowsReturn,&countReturn);

  // If no property, just create one
  if(!status){
    windows[0]=id();
    windows[1]=getShell()->id();
    XSetWMColormapWindows(getDisplay(),getShell()->id(),windows,2);
    }

  // There was a property, add myself to the beginning
  else{
    windowList=(Window*)malloc((sizeof(Window))*(countReturn+1));
    windowList[0]=id();
    for(int i=0; i<countReturn; i++) windowList[i+1]=windowsReturn[i];
    XSetWMColormapWindows(getDisplay(),getShell()->id(),windowList,countReturn+1);
    XFree((char*)windowsReturn);
    free(windowList);
    }
  }


// Remove it from colormap windows
void FXWindow::removeColormapWindows(){
  Window *windowsReturn;
  int countReturn;
  Status status=XGetWMColormapWindows(getDisplay(),getShell()->id(),&windowsReturn,&countReturn);
  if(status){
    for(int i=0; i<countReturn; i++){
      if(windowsReturn[i]==id()){
        for(i++; i<countReturn; i++) windowsReturn[i-1]=windowsReturn[i];
        XSetWMColormapWindows(getDisplay(),getShell()->id(),windowsReturn,countReturn-1);
        break;
        }
      }
    XFree((char*)windowsReturn);
    }
  }

#endif

/*******************************************************************************/


// Set cursor
void FXWindow::setDefaultCursor(FXCursor* cur){
  if(defaultCursor!=cur){
    if(cur==NULL){ fxerror("%s::setDefaultCursor: NULL cursor argument.\n",getClassName()); }
    if(xid){
      if(cur->id()==0){ fxerror("%s::setDefaultCursor: Cursor has not been created yet.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
      XDefineCursor(getDisplay(),xid,cur->id());
#else
      SetCursor((HCURSOR)cur->id());
#endif
      }
    defaultCursor=cur;
    }
  }


// Set drag cursor
void FXWindow::setDragCursor(FXCursor* cur){
  if(dragCursor!=cur){
    if(cur==NULL){ fxerror("%s::setDragCursor: NULL cursor argument.\n",getClassName()); }
    if(xid){
      if(cur->id()==0){ fxerror("%s::setDragCursor: Cursor has not been created yet.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
      if(grabbed()){ XChangeActivePointerGrab(getDisplay(),GRAB_EVENT_MASK,cur->id(),CurrentTime); }
#else
    //if(grabbed()) SetCursor((HCURSOR)cur->id());
#endif
      }
    dragCursor=cur;
    }
  }
  

// Set window background
void FXWindow::setBackColor(FXColor clr){
  if(clr!=backColor){
    backColor=clr;
    update(0,0,width,height);
    }
  }


/*******************************************************************************/


// Has this window the selection
FXbool FXWindow::hasSelection() const {
  return (flags&FLAG_SELECTION)!=0;
  }


// Lost the selection
long FXWindow::onSelectionLost(FXObject*,FXSelector,void* ptr){
  FXTRACE((200,"%s::onSelectionLost %x\n",getClassName(),this));
  flags&=~FLAG_SELECTION;
  if(target && target->handle(this,MKUINT(message,SEL_SELECTION_LOST),ptr)) return 1;
  return 1;
  }


// Gained the selection
long FXWindow::onSelectionGained(FXObject*,FXSelector,void* ptr){
  FXTRACE((200,"%s::onSelectionGained %x\n",getClassName(),this));
  flags|=FLAG_SELECTION;
  if(target && target->handle(this,MKUINT(message,SEL_SELECTION_GAINED),ptr)) return 1;
  return 1;
  }


// Somebody wants our the selection
long FXWindow::onSelectionRequest(FXObject*,FXSelector,void* ptr){
  FXTRACE((200,"%s::onSelectionRequest %x\n",getClassName(),this));
  if(target && target->handle(this,MKUINT(message,SEL_SELECTION_REQUEST),ptr)) return 1;
  return 0;
  }


// Acquire the selection
void FXWindow::acquireSelection(){
  if(!(flags&FLAG_SELECTION)){
    if(xid){
#ifndef FX_NATIVE_WIN32
      
      // A bit tricky, but we need to force a selection-lost on whatever
      // widget held the selection first [Hence two calls, instead of one!]
      XSetSelectionOwner(getDisplay(),XA_PRIMARY,None,getApp()->event.time);
      XSetSelectionOwner(getDisplay(),XA_PRIMARY,xid,getApp()->event.time);
      
      // If we actually got the selection now, post a selection-gained
      if(XGetSelectionOwner(getDisplay(),XA_PRIMARY)==xid){
        handle(this,MKUINT(0,SEL_SELECTION_GAINED),&getApp()->event);
        }
#else
      
      // MSWindows does not have the concept of a selection like X11
      // We do need to post a selection-lost so as to select the text
      handle(this,MKUINT(0,SEL_SELECTION_GAINED),&getApp()->event);
#endif
      }
    }
  }


// Release the selection
void FXWindow::releaseSelection(){
  if(flags&FLAG_SELECTION){
    if(xid){ 
#ifndef FX_NATIVE_WIN32
      
      // In X, this will cause a selection-lost event later on
      if(XGetSelectionOwner(getDisplay(),XA_PRIMARY)==xid){
        XSetSelectionOwner(getDisplay(),XA_PRIMARY,None,CurrentTime);
        }
#else
      
      // MSWindows does not have the concept of a selection like X11
      // We do need to post a selection-lost so as to unselect the text
      handle(this,MKUINT(0,SEL_SELECTION_LOST),&getApp()->event);
#endif
      }
    }
  }


/*******************************************************************************/


// Make transient for some other window
void FXWindow::setTransientFor(FXWindow* main){
  if(xid && main && main->id()){
#ifndef FX_NATIVE_WIN32
    XSetTransientForHint(getDisplay(),xid,main->id());
#else
    ////// FIXME: transient for means stay on top of in X11
    SetWindowPos((HWND)main->id(),(HWND)xid,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
#endif
    }
  }


// Get pointer location (in window coordinates)
FXint FXWindow::getCursorPosition(FXint& x,FXint& y,FXuint& buttons) const {
  if(xid!=0){
#ifndef FX_NATIVE_WIN32
    Window dum; int rx,ry;
    return XQueryPointer(getDisplay(),xid,&dum,&dum,&rx,&ry,&x,&y,&buttons);
#else
    POINT pt;
    GetCursorPos(&pt);
    ScreenToClient((HWND)xid,&pt);
    x=pt.x; y=pt.y;
    buttons=fxmodifierkeys();
    return TRUE;
#endif
    }
  return FALSE;
  }


// Recalculate layout if marked as dirty during some previous interaction.
// This one should ALWAYS be called, as it it responsible for widget layout!
// If updating is not blocked [e.g. during a user-interaction], call upon
// the target to update this widget also.
// Return 0 if we're updatable but no target was present or the target
// didn't handle the update message; we need this for auto-gray or auto-hide
// capability.
long FXWindow::onUpdate(FXObject*,FXSelector,void*){
  if(flags&FLAG_DIRTY) layout();
  if(flags&FLAG_UPDATE){
    FXTRACE((400,"%s::onUpdate %08x\n",getClassName(),this));
    return target && target->handle(this,MKUINT(message,SEL_UPDATE),NULL);
    }
  return 1;
  }


// Propagate window size change
void FXWindow::layout(){ 
  flags&=~FLAG_DIRTY;
  }


// Propagate size change upward
// We also schedule an eventual refresh
void FXWindow::recalc(){
  getApp()->refresh();
  flags|=FLAG_DIRTY;
  parent->recalc(); 
  }


// Update dirty rectangle
void FXWindow::update(FXint x,FXint y,FXint w,FXint h){
  if(xid){

    // Probably got an update before window was laid out; we correct by 
    // building a big expose rectangle which covers the whole window.
    if(w<=0) w=4096;
    if(h<=0) h=4096;

    // Append the rectangle; it is a synthetic expose event!!
#ifndef FX_NATIVE_WIN32
    getApp()->addRepaint(xid,x,y,w,h,1);
#else
    RECT r;
    r.left=x;
    r.top=y;
    r.right=x+w;
    r.bottom=y+h;
    InvalidateRect((HWND)xid,&r,TRUE);
#endif
    }
  }


// Move and resize
void FXWindow::position(FXint x,FXint y,FXint w,FXint h){
  if((flags&FLAG_DIRTY) || (x!=xpos) || (y!=ypos) || (w!=width) || (h!=height)){
    xpos=x;
    ypos=y;
    width=FXMAX(w,1);
    height=FXMAX(h,1);
    if(xid){
#ifndef FX_NATIVE_WIN32
      XMoveResizeWindow(getDisplay(),xid,xpos,ypos,width,height);
#else
      SetWindowPos((HWND)xid,NULL,xpos,ypos,width,height,SWP_NOZORDER);
#endif
      layout();
      }
    }
  }


// Move window
void FXWindow::move(int x,int y){
  if((x!=xpos) || (y!=ypos)){
    xpos=x;
    ypos=y;
    if(xid){
#ifndef FX_NATIVE_WIN32
      XMoveWindow(getDisplay(),xid,xpos,ypos);
#else
      SetWindowPos((HWND)xid,NULL,xpos,ypos,0,0,SWP_NOSIZE|SWP_NOZORDER);
#endif
      }
    }
  }


// Resize
void FXWindow::resize(int w,int h) {
  if((flags&FLAG_DIRTY) || (w!=width) || (h!=height)){
    width=FXMAX(w,1);
    height=FXMAX(h,1);
    if(xid){
#ifndef FX_NATIVE_WIN32
      XResizeWindow(getDisplay(),xid,width,height);
#else
      SetWindowPos((HWND)xid,NULL,0,0,width,height,SWP_NOMOVE|SWP_NOZORDER);
#endif
      layout();
      }
    }
  }


// Reparent window under a new parent
void FXWindow::reparent(FXComposite* newparent){
  if(newparent==NULL){ fxerror("%s::reparent: NULL parent specified.\n",getClassName()); }
  if(parent==NULL){ fxerror("%s::reparent: cannot reparent root window.\n",getClassName()); }
  if(newparent!=parent){
    
    // Check for funny cases
    if(containsChild(newparent)){ fxerror("%s::reparent: new parent is child of window.\n",getClassName()); }
    
    // Both windows created or both non-created
    if(xid && !newparent->xid){ fxerror("%s::reparent: new parent not created yet.\n",getClassName()); }
    if(!xid && newparent->xid){ fxerror("%s::reparent: window not created yet.\n",getClassName()); }
    
    // Kill focus chain through this window
    killFocus();
    
    // Flag old parent as to be recalculated
    parent->recalc();
    
    // Unlink from old parent
    if(prev) prev->next=next; else if(parent) parent->first=next;
    if(next) next->prev=prev; else if(parent) parent->last=prev;
    
    // Link to new parent
    parent=newparent;
    prev=parent->last;
    next=NULL;
    parent->last=this;
    if(prev) prev->next=this; else parent->first=this;
    
    // New toplevel shell too
    shell=parent->getShell();
    
    // Hook up to new window in server too
    if(xid && parent->xid){
#ifndef FX_NATIVE_WIN32
      XReparentWindow(getDisplay(),xid,parent->xid,0,0);
#else
      SetParent((HWND)xid,(HWND)parent->xid);
#endif
      }
    
    // Flag as to be recalculated
    recalc();
    }
  }


// Move window in chain before a sibling
void FXWindow::linkBefore(FXWindow* sibling){
  if(!sibling){fxerror("%s::linkBefore: NULL argument.\n",getClassName());}
  if(sibling->parent!=parent){fxerror("%s::linkBefore: windows should have same parent.\n",getClassName());}
  if(sibling==this){fxerror("%s::linkBefore: same window.\n",getClassName());}
  if(prev) prev->next=next; else if(parent) parent->first=next;
  if(next) next->prev=prev; else if(parent) parent->last=prev; 
  prev=sibling->prev;
  next=sibling;
  if(prev) prev->next=this; else parent->first=this;
  sibling->prev=this;
  recalc();
  }


// Move window in chain after a sibling
void FXWindow::linkAfter(FXWindow* sibling){
  if(!sibling){fxerror("%s::linkAfter: NULL argument.\n",getClassName());}
  if(sibling->parent!=parent){fxerror("%s::linkAfter: windows should have same parent.\n",getClassName());}
  if(sibling==this){fxerror("%s::linkAfter: same window.\n",getClassName());}
  if(prev) prev->next=next; else if(parent) parent->first=next;
  if(next) next->prev=prev; else if(parent) parent->last=prev; 
  next=sibling->next;
  prev=sibling;
  if(next) next->prev=this; else parent->last=this;
  sibling->next=this;
  recalc();
  }


// Map window
void FXWindow::show(){
  if(!(flags&FLAG_SHOWN)){
    if(xid){
#ifndef FX_NATIVE_WIN32
      XMapWindow(getDisplay(),xid);
#else
      ShowWindow((HWND)xid,SW_NORMAL);
#endif
      flags|=FLAG_SHOWN;
      }
    }
  }


// Unmap
void FXWindow::hide(){
  if(flags&FLAG_SHOWN){
    if(xid){
#ifndef FX_NATIVE_WIN32
      XUnmapWindow(getDisplay(),xid);
#else
      ShowWindow((HWND)xid,SW_HIDE);
#endif
      flags&=~FLAG_SHOWN;
      }
    }
  }


// Check if mapped
FXbool FXWindow::shown() const { 
  return (flags&FLAG_SHOWN)!=0; 
  }


// Test if logically inside
FXbool FXWindow::contains(FXint parentx,FXint parenty) const { 
  return xpos<=parentx && parentx<xpos+width && ypos<=parenty && parenty<ypos+height;
  }


// Enable the window
void FXWindow::enable(){
  if(!(flags&FLAG_ENABLED)){
    if(xid){
#ifndef FX_NATIVE_WIN32
      FXuint events=BASIC_EVENT_MASK|ENABLED_EVENT_MASK;
      if(shell==this) events|=SHELL_EVENT_MASK;
      XSelectInput(getDisplay(),xid,events);
#else
      EnableWindow((HWND)xid,TRUE);
#endif
      }
    flags|=FLAG_ENABLED;
    }
  }


// Disable the window
void FXWindow::disable(){ 
  if(flags&FLAG_ENABLED){
    if(xid){
#ifndef FX_NATIVE_WIN32
      FXuint events=BASIC_EVENT_MASK;
      if(shell==this) events|=SHELL_EVENT_MASK;
      XSelectInput(getDisplay(),xid,events);
#else
      EnableWindow((HWND)xid,FALSE);
#endif
      }
    flags&=~FLAG_ENABLED;
    }
  }


// Is window enabled
FXbool FXWindow::isEnabled() const { 
  return (flags&FLAG_ENABLED)!=0; 
  }


// Raise
void FXWindow::raise(){ 
  if(xid){ 
#ifndef FX_NATIVE_WIN32
    XRaiseWindow(getDisplay(),xid); 
#else
    BringWindowToTop((HWND)xid);
#endif
    }
  }
  

// Lower
void FXWindow::lower(){ 
  if(xid){ 
#ifndef FX_NATIVE_WIN32
    XLowerWindow(getDisplay(),xid); 
#else
    SetWindowPos((HWND)xid,HWND_BOTTOM,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
#endif
    }
  }



// Locate window at root coordinate position x,y
FXID FXWindow::getWindowAt(FXint x,FXint y) const {
#ifndef FX_NATIVE_WIN32
  Window win,child=XDefaultRootWindow(getDisplay());
  int dropx,dropy;
  while(1){
    if(!XTranslateCoordinates(getDisplay(),XDefaultRootWindow(getDisplay()),child,x,y,&dropx,&dropy,&win)){
      return XDefaultRootWindow(getDisplay());
      }
    if(win==None) break;
    child=win;
    }
  return child;
#else
  POINT point;
  point.x=x; point.y=y;
  return WindowFromPoint(point);
#endif
  }


// Get coordinates from another window (for symmetry)
void FXWindow::translateCoordinatesFrom(FXint& tox,FXint& toy,const FXWindow* fromwindow,FXint fromx,FXint fromy) const {
  if(fromwindow==NULL){ fxerror("%s::translateCoordinatesFrom: from-window is NULL.\n",getClassName()); }
  if(xid && fromwindow->xid){
#ifndef FX_NATIVE_WIN32
    Window tmp;
    XTranslateCoordinates(getDisplay(),fromwindow->xid,xid,fromx,fromy,&tox,&toy,&tmp);
#else
    POINT pt;
    pt.x=fromx; pt.y=fromy;
    ClientToScreen((HWND)fromwindow->xid,&pt);
    ScreenToClient((HWND)xid,&pt);
    tox=pt.x; toy=pt.y;
#endif
    }
  }


// Get coordinates to another window (for symmetry)
void FXWindow::translateCoordinatesTo(FXint& tox,FXint& toy,const FXWindow* towindow,FXint fromx,FXint fromy) const {
  if(towindow==NULL){ fxerror("%s::translateCoordinatesTo: to-window is NULL.\n",getClassName()); }
  if(xid && towindow->xid){
#ifndef FX_NATIVE_WIN32
    Window tmp;
    XTranslateCoordinates(getDisplay(),xid,towindow->xid,fromx,fromy,&tox,&toy,&tmp);
#else
    POINT pt;
    pt.x=fromx; pt.y=fromy;
    ClientToScreen((HWND)xid,&pt);
    ScreenToClient((HWND)towindow->xid,&pt);
    tox=pt.x; toy=pt.y;
#endif
    }
  }


/*******************************************************************************/


// Acquire grab
// Grabs also switch to the drag cursor
void FXWindow::grab(){
  if(xid){
    FXTRACE((150,"%s::grab\n",getClassName()));
    if(dragCursor->id()==0){ fxerror("%s::grab: Cursor has not been created yet.\n",getClassName()); }
    if(!(flags&FLAG_SHOWN)){ fxwarning("%s::grab: Window is not visible.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
    if(GrabSuccess!=XGrabPointer(getDisplay(),xid,FALSE,GRAB_EVENT_MASK,GrabModeAsync,GrabModeAsync,None,dragCursor->id(),CurrentTime)){ fxwarning("%s::grab: grab failed!\n",getClassName()); }
#else
    SetCapture((HWND)xid);
    if(GetCapture()!=(HWND)xid){ fxwarning("%s::grab: grab failed!\n",getClassName()); }
#endif
    getApp()->grabWindow=this;
    }
  }


// Release grab
// Ungrabs also switch back to the normal cursor
void FXWindow::ungrab(){
  if(xid){
    FXTRACE((150,"%s::ungrab\n",getClassName()));
#ifndef FX_NATIVE_WIN32
    XUngrabPointer(getDisplay(),CurrentTime);
    XFlush(getDisplay());
#else
    ReleaseCapture();
#endif
    getApp()->grabWindow=NULL;
    }
  }


// Return true if active grab is in effect
FXbool FXWindow::grabbed() const {
  return getApp()->grabWindow==this;
  }


// The widget lost the grab for some reason (Windows)
// Get this message when system forced ungrab on us
// Subclasses should try to clean up the mess...
long FXWindow::onUngrabbed(FXObject*,FXSelector,void* ptr){
//FXTRACE((10,"%s::ungrab forced\n",getClassName()));
  FXASSERT(getApp()->grabWindow==this);
  getApp()->grabWindow=NULL;
  if(isEnabled()){
    if(target && target->handle(this,MKUINT(message,SEL_UNGRABBED),ptr)) return 1;
    }
  return 0;
  }


/*******************************************************************************/

// Acquire the clipboard
FXbool FXWindow::takeClipboard(){
  if(xid){
#ifndef FX_NATIVE_WIN32
    
    // Got it already
    if(XGetSelectionOwner(getDisplay(),getApp()->cbSelection)==xid) return TRUE;
      
    // A bit tricky, but we need to force a selection-lost on whatever
    // widget held the selection first [Hence two calls, instead of one!]
    XSetSelectionOwner(getDisplay(),getApp()->cbSelection,None,getApp()->event.time);
    XSetSelectionOwner(getDisplay(),getApp()->cbSelection,xid,getApp()->event.time);
    
    // If we actually got the selection now, post a selection-gained
    if(XGetSelectionOwner(getDisplay(),getApp()->cbSelection)==xid) return TRUE;

#else

    // Under windows, we just claim the clipboard always with CF_TEXT and a few others.
    // Also, a special FOX type called targets which we can use for a runtime-determined
    // list of clipboard types.
    if(OpenClipboard((HWND)xid)){
      EmptyClipboard();
      SetClipboardData(CF_TEXT,NULL);
      CloseClipboard();
      }

#endif
    }
  return FALSE;
  }


/*******************************************************************************/

#ifndef FX_NATIVE_WIN32

// To pass arguments to matcher
struct XMatch {
  Window window;
  Atom   selection;
  int    gotit;
  };

#endif

// Enable widget as drop target
void FXWindow::dropEnable(){
  if(xid==0){ fxerror("%s::dropEnable: window has not yet been created.\n",getClassName()); }
  if(!(flags&FLAG_DROPTARGET)){
#ifndef FX_NATIVE_WIN32
    FXID propdata=(FXID)XDND_PROTOCOL_VERSION;
    FXASSERT(getApp()->xdndAware);
    XChangeProperty(getDisplay(),xid,getApp()->xdndAware,XA_ATOM,32,PropModeReplace,(FXuchar*)&propdata,1);
#else
    FXASSERT(FALSE);
#endif
    flags|=FLAG_DROPTARGET;
    }
  }


// Disable widget as drop target
void FXWindow::dropDisable(){
  if(xid==0){ fxerror("%s::dropDisable: window has not yet been created.\n",getClassName()); }
  if(flags&FLAG_DROPTARGET){
#ifndef FX_NATIVE_WIN32
    FXASSERT(getApp()->xdndAware);
    XDeleteProperty(getDisplay(),xid,getApp()->xdndAware);
#else
    FXASSERT(FALSE);
#endif
    flags&=~FLAG_DROPTARGET;
    }
  }


// Is window drop enabled
FXbool FXWindow::isDropEnabled() const {
  return (flags&FLAG_DROPTARGET)!=0;
  }


// Set drag rectangle; we'll get no more DND_MOTION events while cursor is inside
void FXWindow::setDragRectangle(FXint x,FXint y,FXint w,FXint h,FXbool wantupdates){
  if(xid==0){ fxerror("%s::setDragRectangle: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  int tox,toy;
  Window tmp;
  XTranslateCoordinates(getDisplay(),parent->id(),XDefaultRootWindow(getDisplay()),x,y,&tox,&toy,&tmp);
  getApp()->xdndRect.x=tox;
  getApp()->xdndRect.y=toy;
  getApp()->xdndRect.w=w;
  getApp()->xdndRect.h=h;
  getApp()->xdndWantUpdates=wantupdates;
#else
  FXASSERT(FALSE);
#endif
  }


// Clear drag rectangle
// Target always wants SEL_DND_MOTION messages
void FXWindow::clearDragRectangle(){
  if(xid==0){ fxerror("%s::clearDragRectangle: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  getApp()->xdndRect.x=0;
  getApp()->xdndRect.y=0;
  getApp()->xdndRect.w=0;
  getApp()->xdndRect.h=0;
  getApp()->xdndWantUpdates=FALSE;
#else
  FXASSERT(FALSE);
#endif
  }


// Wait for arrival of dynamically exchanged data
#ifndef FX_NATIVE_WIN32
static Bool notifyevent(Display *,XEvent *event,XPointer ptr){
  XMatch *match=(XMatch*)ptr;
  if(event->type==DestroyNotify && event->xdestroywindow.window==match->window){
    match->gotit=0;
    return True;
    }
  if(event->type==SelectionNotify && event->xselection.selection==match->selection){
    return True;
    }
  return False;
  }
#endif



// Get dropped data; called in response to DND enter or DND drop
FXbool FXWindow::getDNDData(FXDNDOrigin origin,FXDragType targettype,FXuchar*& data,FXuint& size){
#ifndef FX_NATIVE_WIN32
  unsigned long  nitems,alloc_incr,bytes_after,transfersize,offset;
  FXWindow      *ownerobject;
  Window         ownerwindow;
  XEvent         ev;
  XMatch         match;
  Bool           status;
  Atom           actualtype;
  Atom           selection;
  int            actualformat;
  FXuchar       *ptr;

  if(xid==0){ fxerror("%s::getDNDData: window has not yet been created.\n",getClassName()); }

  data=NULL;
  size=0;
  
  // Selection where data comes from
  if(origin==FROM_SELECTION) 
    selection=XA_PRIMARY;
  else if(origin==FROM_CLIPBOARD) 
    selection=getApp()->cbSelection;
  else if(origin==FROM_DRAGNDROP) 
    selection=getApp()->xdndSelection;
  else 
    return FALSE;
  
  // If no owner, we're done quickly
  ownerwindow=XGetSelectionOwner(getDisplay(),selection);
  if(ownerwindow == None) return FALSE;
  
  // Window is in same application; bypass protocol by returning data from the mailbox
  if(XFindContext(getDisplay(),ownerwindow,getApp()->wcontext,(XPointer*)&ownerobject)==0){
    FXASSERT(ownerobject);
    getApp()->event.type=SEL_SELECTION_REQUEST;
    getApp()->event.window=ownerwindow;
    getApp()->event.requestor=xid;
    getApp()->event.target=targettype;
    getApp()->event.property=getApp()->ddeAtom;
    getApp()->event.origin=origin;
    ownerobject->handle(getApp(),MKUINT(0,SEL_SELECTION_REQUEST),&getApp()->event);
    data=getApp()->ddeData;
    size=getApp()->ddeSize;
    getApp()->ddeData=NULL;
    getApp()->ddeSize=0;
    FXTRACE((100,"%s: getDNDData(target=%d,data=%08x,size=%d) from %d (%s)\n",getClassName(),targettype,data,size,ownerwindow,ownerobject->getClassName()));
    return data!=NULL;
    }
  
  // Send message to the selection owner
  XConvertSelection(getDisplay(),selection,targettype,getApp()->ddeAtom,xid,getApp()->event.time);
  
  // Wait for a response
  match.window=xid;
  match.selection=selection;
  match.gotit=1;
  status=XCheckIfEvent(getDisplay(),&ev,notifyevent,(XPointer)&match);

  // Broke out of event loop
  if(!match.gotit) return FALSE;
  if(!status){
    match.gotit=1;
    XIfEvent(getDisplay(),&ev,notifyevent,(XPointer)&match);
    if(!match.gotit) return FALSE;
    }
  
  // There was no selection or no data has been specified
  if(ev.xselection.property==None){
    return FALSE;
    }

  // Prepare for download of big property data
  transfersize=XMaxRequestSize(getDisplay());
  if(!FXMALLOC(&getApp()->ddeData,FXuchar,8)) return FALSE;

  // Loop while we didn't get it all
  offset=0;
  getApp()->ddeSize=0;
  bytes_after=1;
  while(bytes_after){
    if(XGetWindowProperty(getDisplay(),xid,getApp()->ddeAtom,offset,transfersize,False,AnyPropertyType,&actualtype,&actualformat,&nitems,&bytes_after,&ptr)!=Success){
      FXFREE(&getApp()->ddeData);
      getApp()->ddeData=NULL;
      getApp()->ddeSize=0;
      FXTRACE((100,"%s: getDNDData(target=%d) from %d (remote) failed\n",getClassName(),targettype,ownerwindow));
      return FALSE;
      }
    if(ptr == NULL || nitems == 0){
      XDeleteProperty(getDisplay(),xid,getApp()->ddeAtom);
      if(ptr) XFree(ptr);
      FXFREE(&getApp()->ddeData);
      getApp()->ddeData=NULL;
      getApp()->ddeSize=0;
      FXTRACE((100,"%s: getDNDData(target=%d,data=NULL,size=0) from %d (remote)\n",getClassName(),targettype,ownerwindow));
      return FALSE;
      }
    alloc_incr = (actualformat==32) ? nitems<<2 : (actualformat==16) ? nitems<<1 : nitems;
    if(!FXRESIZE(&getApp()->ddeData,FXuchar,getApp()->ddeSize+alloc_incr+1)){
      XDeleteProperty(getDisplay(),xid,getApp()->ddeAtom);
      XFree(ptr);
      getApp()->ddeData=NULL;
      getApp()->ddeSize=0;
      FXTRACE((100,"%s: getDNDData(target=%d) from %d (remote) failed\n",getClassName(),targettype,ownerwindow));
      return FALSE;
      }
    memcpy(&getApp()->ddeData[offset],ptr,alloc_incr);
    getApp()->ddeSize += alloc_incr;
    offset += (actualformat==32) ? nitems : (actualformat==16) ? nitems>>1 : nitems>>2;
    XFree(ptr);
    }
  XDeleteProperty(getDisplay(),xid,getApp()->ddeAtom);
  getApp()->ddeData[getApp()->ddeSize]=0;
  data=getApp()->ddeData;
  size=getApp()->ddeSize;
  getApp()->ddeData=NULL;
  getApp()->ddeSize=0;
  FXTRACE((100,"%s: getDNDData(target=%d,data=%08x,size=%d) from %d (remote)\n",getClassName(),targettype,data,size,ownerwindow));
  return TRUE;

#else

  data=NULL;
  size=0;
  
  // Data from clip board
  if(origin==FROM_CLIPBOARD){
    if(IsClipboardFormatAvailable(targettype)){
      if(OpenClipboard((HWND)xid)){
	HANDLE hClipMemory=GetClipboardData(targettype);
	FXASSERT(hClipMemory);
	if(hClipMemory){
	  size=GlobalSize(hClipMemory);
	  if(FXMALLOC(&data,FXuchar,size)){
	    void *pClipMemory=GlobalLock(hClipMemory);
 	    FXASSERT(pClipMemory);
	    memcpy((void*)data,pClipMemory,size);
	    GlobalUnlock(hClipMemory);
	    CloseClipboard();
	    getApp()->ddeData=NULL;
	    getApp()->ddeSize=0;
	    return TRUE;
	    }
	  else{
	    data=NULL;
	    size=0;
	    }
	  }
	}
      }
    }
  
  // Data from drag and drop
  else if(origin==FROM_DRAGNDROP){
    // Drag and drop not implemented yet
    FXASSERT(FALSE);
    }
  
  // Data from selection not available for MS-Windows
  return FALSE;

#endif
  }



// Set drop data; called in response to data request
// Data array will be deleted by the system automatically!!
FXbool FXWindow::setDNDData(FXDNDOrigin origin,FXDragType targettype,FXuchar* data,FXuint size){
#ifndef FX_NATIVE_WIN32
  unsigned long  transfersize,offset;
  unsigned long  nunits,tlen,mode;
  FXWindow      *reqobject;
  int            format;
  
  if(xid==0){ fxerror("%s::setDNDData: window has not yet been created.\n",getClassName()); }
  if(origin!=getApp()->event.origin){ fxerror("%s::setDNDData: DDE origin was not requested.\n",getClassName()); }

  
  // Window is in same application; bypass protocol by storing data into the mailbox
  if(XFindContext(getDisplay(),getApp()->event.requestor,getApp()->wcontext,(XPointer*)&reqobject)==0){
    FXTRACE((100,"%s: setDNDData(target=%d,data=%08x,size=%d) for %d (%s)\n",getClassName(),targettype,data,size,getApp()->event.requestor,reqobject->getClassName()));
    getApp()->ddeData=data;
    getApp()->ddeSize=size;
    }
  
  // Window in another application.  Stick the data into the indicated property,
  // then delete our local copy.  A new data array will be created in the remote
  // application.
  else{
    FXTRACE((100,"%s: setDNDData(target=%d,data=%08x,size=%d) for %d (remote)\n",getClassName(),targettype,data,size,getApp()->event.requestor));
    if(data!=NULL && size!=0){
      transfersize=XMaxRequestSize(getDisplay());
      mode=PropModeReplace;
      nunits=transfersize<<2;
      offset=0;
      format=8;
      if(targettype==targetType) format=32;   // TARGETS is the request with tfr size of 32
      FXASSERT(size!=0);
      do{
        if(size<nunits) tlen=size; else tlen=nunits;
        XChangeProperty(getDisplay(),getApp()->event.requestor,getApp()->event.property,getApp()->event.target,format,mode,&data[offset],tlen);
        size-=tlen;
        mode=PropModeAppend;
        offset+=transfersize;
        }
      while(size!=0);
      FXFREE(&data);
      }
    else{
      getApp()->event.property=None;
      }
    }
  return TRUE;
  
#else
  
  if(xid==0){ fxerror("%s::setDNDData: window has not yet been created.\n",getClassName()); }
  if(origin!=getApp()->event.origin){ fxerror("%s::setDNDData: DDE source was not requested.\n",getClassName()); }

  FXTRACE((100,"%s: setDNDData()\n",getClassName()));

  // Was called in response to a WM_RENDERFORMAT
  if(origin==FROM_CLIPBOARD){
    HGLOBAL hGlobalMemory=GlobalAlloc(GHND,size);
    FXASSERT(hGlobalMemory);
    if(hGlobalMemory){
      void *pGlobalMemory=GlobalLock(hGlobalMemory);
      FXASSERT(pGlobalMemory);
      memcpy((FXchar*)pGlobalMemory,data,size);
      GlobalUnlock(hGlobalMemory);
      SetClipboardData(targettype,hGlobalMemory);
      FXFREE(&getApp()->ddeData);
      getApp()->ddeData=NULL;
      getApp()->ddeSize=0;
      return TRUE;
      }
    return FALSE;
    }
  FXASSERT(FALSE);
  return FALSE;
#endif
  }


// Target says accepts the drop
void FXWindow::acceptDrop(FXDragAction action){
  if(xid==0){ fxerror("%s::acceptDrop: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  if(getApp()->xdndTarget!=xid){ fxerror("%s::acceptDrop: window is not the drop target.\n",getClassName()); }

  // Accept anything or suggest a specific action?
  if(action!=DRAG_ACCEPT){
    if(action==DRAG_COPY) getApp()->xdndAction=getApp()->xdndActionCopy;
    if(action==DRAG_MOVE) getApp()->xdndAction=getApp()->xdndActionMove;
    if(action==DRAG_LINK) getApp()->xdndAction=getApp()->xdndActionLink;
    }
  
  // Accept a drop?
  getApp()->xdndAccepts=(action!=DRAG_REJECT);
#else
  FXASSERT(FALSE);
#endif
  }



// Inquire about types being dragged over this window
FXbool FXWindow::inquireDNDTypes(const FXDragType*& types,FXuint& numtypes) const {
  if(xid==0){ fxerror("%s::inquireDNDTypes: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  if(getApp()->xdndTarget!=xid){ fxerror("%s::inquireDNDTypes: window is not the drop target.\n",getClassName()); }
  types=NULL;
  numtypes=0;
  if(getApp()->xdndNumTypes){
    types=getApp()->xdndTypeList;
    numtypes=getApp()->xdndNumTypes;
    return TRUE;
    }
#else
  FXASSERT(FALSE);
#endif
  return FALSE;
  }


// Is the drop target being offered a certain DND type?
FXbool FXWindow::offeredDNDType(FXDragType type) const {
  if(xid==0){ fxerror("%s::offeredDNDType: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  if(getApp()->xdndTarget!=xid){ fxerror("%s::offeredDNDType: window is not the drop target.\n",getClassName()); }
  for(FXuint i=0; i<getApp()->xdndNumTypes; i++){
    if(type==getApp()->xdndTypeList[i]) return TRUE;
    }
#else
  FXASSERT(FALSE);
#endif
  return FALSE;
  }


// What is the drag action being performed?
FXDragAction FXWindow::inquireDNDAction() const {
  FXDragAction action=DRAG_COPY;
  if(xid==0){ fxerror("%s::inquireDNDAction: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  if(getApp()->xdndTarget!=xid){ fxerror("%s::inquireDNDAction: window is not the drop target.\n",getClassName()); }
  if(getApp()->xdndAction==getApp()->xdndActionCopy) action=DRAG_COPY;
  if(getApp()->xdndAction==getApp()->xdndActionMove) action=DRAG_MOVE;
  if(getApp()->xdndAction==getApp()->xdndActionLink) action=DRAG_LINK;
#else
  FXASSERT(FALSE);
#endif
  return action;
  }


// Return non-zero if we can drop on window
FXuint FXWindow::mayDropOn(FXID dropwindow) const {
#ifndef FX_NATIVE_WIN32
  unsigned long itemCount,remainingBytes;
  unsigned char* rawData=NULL;
  Atom actualType;
  FXuint version=0;
  FXuint i,j;
  int actualFormat;
  if(XGetWindowProperty(getDisplay(),dropwindow,getApp()->xdndAware,0,1024,False,XA_ATOM,&actualType,&actualFormat,&itemCount,&remainingBytes,&rawData)==Success){
    if(rawData){
      if(actualType==XA_ATOM && actualFormat==32 && itemCount>0){
        
        // Version of the protocol
        version=*((FXuint*)rawData) | 0x80000000;
        
        // Did target specify specific types?
        if(itemCount>1){
          
          // Try match up the type lists
          for(i=0; i<getApp()->xdndNumTypes; i++){
            for(j=1; j<itemCount; j++){
              if(getApp()->xdndTypeList[i]==((FXuint*)rawData)[j]) goto ok;
              }
            }
          
          // Too bad, didn't find it
          version=0;
          }
        }
ok:   XFree(rawData);
      }
    }
  return version;
#else
  FXASSERT(FALSE);
  return 0;
#endif
  }



// Source wants to find out if target accepted
FXDragAction FXWindow::didAccept() const {
  FXDragAction action=DRAG_REJECT;
  if(xid==0){ fxerror("%s::didAccept: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  if(getApp()->xdndSource!=xid){ fxerror("%s::didAccept: window is not a drag source.\n",getClassName()); }
  if(getApp()->xdndAccepts){
    if(getApp()->xdndAction==getApp()->xdndActionCopy) action=DRAG_COPY;
    if(getApp()->xdndAction==getApp()->xdndActionMove) action=DRAG_MOVE;
    if(getApp()->xdndAction==getApp()->xdndActionLink) action=DRAG_LINK;
    }
#else
  FXASSERT(FALSE);
#endif
  return action;
  }


// True if we're in a drag operation
FXbool FXWindow::isDragging() const { 
  return (flags&FLAG_DRAGGING)!=0; 
  }


// True if this window is drop target
FXbool FXWindow::isDropTarget() const {
  if(xid==0){ fxerror("%s::isDropTarget: window has not yet been created.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
  return xid==getApp()->xdndTarget;
#else
  return FALSE;
#endif
  }
  

// Start a drag on the types mentioned
FXbool FXWindow::beginDrag(const FXDragType *types,FXuint numtypes){
  if(xid==0){ fxerror("%s::beginDrag: window has not yet been created.\n",getClassName()); }
  if(!(flags&FLAG_DRAGGING)){
    if(numtypes<1){ fxerror("%s::beginDrag: should have at least one type to drag.\n",getClassName()); }
#ifndef FX_NATIVE_WIN32
    XSetSelectionOwner(getDisplay(),getApp()->xdndSelection,xid,getApp()->event.time);
    if(XGetSelectionOwner(getDisplay(),getApp()->xdndSelection)!=xid){
      fxwarning("%s::beginDrag: failed to acquire DND selection.\n",getClassName());
      return FALSE;
      }
    getApp()->xdndTypeList=(FXID*)malloc(sizeof(FXID)*numtypes);
    if(!getApp()->xdndTypeList) return FALSE;
    getApp()->xdndNumTypes=numtypes;
    for(FXuint num=0; num<numtypes; num++){
      getApp()->xdndTypeList[num]=types[num];
      }
    XChangeProperty(getDisplay(),xid,getApp()->xdndTypes,XA_ATOM,32,PropModeReplace,(const FXuchar*)getApp()->xdndTypeList,getApp()->xdndNumTypes);
    getApp()->xdndSource=xid;
    getApp()->xdndTarget=xid;
    getApp()->xdndAccepts=FALSE;
    getApp()->xdndSendPosition=FALSE;
    getApp()->xdndStatusPending=FALSE;
    getApp()->xdndStatusReceived=FALSE;
    getApp()->xdndWantUpdates=FALSE;
    getApp()->xdndVersion=XDND_PROTOCOL_VERSION;
    getApp()->xdndAction=getApp()->xdndActionCopy;
    getApp()->xdndRect.x=0;
    getApp()->xdndRect.y=0;
    getApp()->xdndRect.w=0;
    getApp()->xdndRect.h=0;
    getApp()->xdndXPos=0;
    getApp()->xdndYPos=0;
#else
    FXASSERT(FALSE);
#endif
    flags|=FLAG_DRAGGING;
    return TRUE;
    }
  return FALSE;
  }


// Drag to new position
FXbool FXWindow::handleDrag(FXint x,FXint y,FXDragAction action){
#ifndef FX_NATIVE_WIN32
  FXuint version=XDND_PROTOCOL_VERSION;
  FXuint forcepos=0;
  FXWindow *tgt;
  Window child,tmp;
  XEvent se;
  if(xid==0){ fxerror("%s::handleDrag: window has not yet been created.\n",getClassName()); }
  if(flags&FLAG_DRAGGING){
    if(getApp()->xdndSource!=xid){ fxerror("%s::handleDrag: window is not a drag source.\n",getClassName()); }
    if(action<DRAG_COPY || DRAG_LINK<action){ fxerror("%s::handleDrag: illegal drag action.\n",getClassName()); }
  
    // Get drop window
    child=getWindowAt(x,y);
    
    // Got proper drop window?
    if(child!=XDefaultRootWindow(getDisplay())){
      
      // Test for XDND awareness
      version=mayDropOn(child);
      if(!version){
        child=XDefaultRootWindow(getDisplay());
        }
      }
    
    version&=0xff;
    
    // Moved to different window?
    if(child!=getApp()->xdndTarget){
      
      // Moving OUT of XDND aware window?
      if(getApp()->xdndTarget!=XDefaultRootWindow(getDisplay())){
        
        // If local, bypass protocol
        if(XFindContext(getDisplay(),getApp()->xdndTarget,getApp()->wcontext,(XPointer*)&tgt)==0){
          FXASSERT(tgt);
          getApp()->event.type=SEL_DND_LEAVE;
          getApp()->event.window=getApp()->xdndTarget;
          tgt->handle(this,MKUINT(0,SEL_DND_LEAVE),&getApp()->event);
          }
        
        // Else send through protocol
        else{
          se.xclient.type=ClientMessage;
          se.xclient.display=getDisplay();
          se.xclient.message_type=getApp()->xdndLeave;
          se.xclient.format=32;
          se.xclient.window=getApp()->xdndTarget;
          se.xclient.data.l[0]=getApp()->xdndSource;
          se.xclient.data.l[1]=0;
          se.xclient.data.l[2]=0;
          se.xclient.data.l[3]=0;
          se.xclient.data.l[4]=0;
          XSendEvent(getDisplay(),getApp()->xdndTarget,True,NoEventMask,&se);
          }
        }
      
      // Reset XDND variables
      getApp()->xdndAccepts=FALSE;
      getApp()->xdndSendPosition=FALSE;
      getApp()->xdndStatusPending=FALSE;
      getApp()->xdndStatusReceived=FALSE;
      getApp()->xdndVersion=FXMIN(version,XDND_PROTOCOL_VERSION);
      getApp()->xdndAction=getApp()->xdndActionCopy;
      getApp()->xdndWantUpdates=FALSE;
      getApp()->xdndRect.x=0;
      getApp()->xdndRect.y=0;
      getApp()->xdndRect.w=0;
      getApp()->xdndRect.h=0;
      getApp()->xdndXPos=0;
      getApp()->xdndYPos=0;
      getApp()->xdndTarget=child;
      
      // Moving INTO XDND aware window?
      if(getApp()->xdndTarget!=XDefaultRootWindow(getDisplay())){
        
        // If local, bypass protocol
        if(XFindContext(getDisplay(),getApp()->xdndTarget,getApp()->wcontext,(XPointer*)&tgt)==0){
          FXASSERT(tgt);
          getApp()->event.type=SEL_DND_ENTER;
          getApp()->event.window=getApp()->xdndTarget;
          tgt->handle(this,MKUINT(0,SEL_DND_ENTER),&getApp()->event);
          }
        
        // Else send through protocol
        else{
          se.xclient.type=ClientMessage;
          se.xclient.display=getDisplay();
          se.xclient.message_type=getApp()->xdndEnter;
          se.xclient.format=32;
          se.xclient.window=getApp()->xdndTarget;
          se.xclient.data.l[0]=getApp()->xdndSource;
          se.xclient.data.l[1]=getApp()->xdndVersion<<24;
          se.xclient.data.l[2]=getApp()->xdndNumTypes>=1 ? getApp()->xdndTypeList[0] : None;
          se.xclient.data.l[3]=getApp()->xdndNumTypes>=2 ? getApp()->xdndTypeList[1] : None;
          se.xclient.data.l[4]=getApp()->xdndNumTypes>=3 ? getApp()->xdndTypeList[2] : None;
          if(getApp()->xdndNumTypes>3) se.xclient.data.l[1]|=1;
          XSendEvent(getDisplay(),getApp()->xdndTarget,True,NoEventMask,&se);
          }
        forcepos=1;
        }
      }

    // Send target a position update
    if(getApp()->xdndTarget!=XDefaultRootWindow(getDisplay())){
      
      // Send position if we're outside the mouse box or ignoring it
      if(forcepos || getApp()->xdndRect.w==0 || getApp()->xdndRect.h==0 || (getApp()->xdndWantUpdates && getApp()->xdndRect.x<=x && getApp()->xdndRect.y<=y && x<getApp()->xdndRect.x+getApp()->xdndRect.w && y<getApp()->xdndRect.y+getApp()->xdndRect.h)){

        // No outstanding status message, so send new pos right away      
        if(!getApp()->xdndStatusPending){
          
          // Update drag action being performed
          if(action==DRAG_COPY) getApp()->xdndAction=getApp()->xdndActionCopy;
          if(action==DRAG_MOVE) getApp()->xdndAction=getApp()->xdndActionMove;
          if(action==DRAG_LINK) getApp()->xdndAction=getApp()->xdndActionLink;
          
          // New position
          getApp()->xdndXPos=x;
          getApp()->xdndYPos=y;
          
          // If local, bypass protocol
          if(XFindContext(getDisplay(),getApp()->xdndTarget,getApp()->wcontext,(XPointer*)&tgt)==0){
            FXASSERT(tgt);
            getApp()->event.type=SEL_DND_MOTION;
            getApp()->event.window=getApp()->xdndTarget;
            getApp()->event.root_x=x;
            getApp()->event.root_y=y;
            XTranslateCoordinates(getDisplay(),XDefaultRootWindow(getDisplay()),getApp()->xdndTarget,x,y,&getApp()->event.win_x,&getApp()->event.win_y,&tmp);
            getApp()->xdndAccepts=FALSE;
            tgt->handle(this,MKUINT(0,SEL_DND_MOTION),&getApp()->event);
            getApp()->xdndStatusReceived=TRUE;
            getApp()->xdndStatusPending=FALSE;      // We know it got there, as its local
            }
          
          // Else send through protocol
          else{
            se.xclient.type=ClientMessage;
            se.xclient.display=getDisplay();
            se.xclient.message_type=getApp()->xdndPosition;
            se.xclient.format=32;
            se.xclient.window=getApp()->xdndTarget;
            se.xclient.data.l[0]=getApp()->xdndSource;
            se.xclient.data.l[1]=0;
            se.xclient.data.l[2]=MKUINT(y,x);                               // Coordinates
            se.xclient.data.l[3]=getApp()->event.time;                      // Time stamp
            se.xclient.data.l[4]=getApp()->xdndAction;                      // Drag and Drop action
            XSendEvent(getDisplay(),getApp()->xdndTarget,True,NoEventMask,&se);
            getApp()->xdndStatusPending=TRUE;       // Waiting for the other app to respond
            }
          }

        // Send it later
        else{
          getApp()->xdndSendPosition=TRUE;
          }
        }
      }
    return TRUE;
    }
#else
  HWND child;
  if(xid==0){ fxerror("%s::handleDrag: window has not yet been created.\n",getClassName()); }
  if(flags&FLAG_DRAGGING){
    if(action<DRAG_COPY || DRAG_LINK<action){ fxerror("%s::handleDrag: illegal drag action.\n",getClassName()); }
  
    // Get drop window
    child=(HWND)getWindowAt(x,y);
    
    // Got proper drop window?
    if(child!=GetDesktopWindow()){
      SendMessage(child,getApp()->xdndEnter,0,0);
      }
    }
#endif
  return FALSE;
  }


#ifndef FX_NATIVE_WIN32

// Scan for status and selection requests...
static Bool matchevent(Display* display,XEvent* event,XPointer match){
  if(event->type == ClientMessage && event->xclient.message_type == *((Atom*)match)) return True;
  if(event->type == SelectionRequest) return True;
  return False;
  }

#endif


// Compare times
static inline int operator<(const struct timeval& a,const struct timeval& b){
  return (a.tv_sec<b.tv_sec) || (a.tv_sec==b.tv_sec && a.tv_usec<b.tv_usec);
  }


// Terminate the drag; if drop flag is false, don't drop even if accepted.
FXbool FXWindow::endDrag(FXbool drop){
  FXbool dropped=FALSE;
#ifndef FX_NATIVE_WIN32
  FXbool nodrop=TRUE;
  FXWindow *tgt;
  XEvent se;
  FXuint loops;
  if(xid==0){ fxerror("%s::endDrag: window has not yet been created.\n",getClassName()); }
  if(flags&FLAG_DRAGGING){
    if(getApp()->xdndSource!=xid){ fxerror("%s::endDrag: window is not a drag source.\n",getClassName()); }

    // Ever received a status
    if(getApp()->xdndStatusReceived && drop){
      
      // If a status message is still pending, wait for it
      if(getApp()->xdndStatusPending){
        getApp()->xdndSendPosition=FALSE;          // No more position updates sent to this target
        loops=1000;
        while(loops){
          if(XCheckIfEvent(getDisplay(),&se,matchevent,(char*)&getApp()->xdndStatus)){
            getApp()->dispatchEvent(se);
            if(se.xany.type==ClientMessage && se.xclient.data.l[0]==getApp()->xdndTarget){
//fprintf(stderr,"Got last status message!\n");
              getApp()->xdndStatusPending=FALSE;
              break;
              }
 
            // We either got a selection request or a status message for a stale target.
            // Since there is thus still hope for a message, lets wait a bit longer...
            loops=1000;
            }
          fxsleep(20000);
          loops--;
          }
        }
      
      // Got our status message
      if(!getApp()->xdndStatusPending && getApp()->xdndAccepts){
        
        // If local, bypass protocol
        if(XFindContext(getDisplay(),getApp()->xdndTarget,getApp()->wcontext,(XPointer*)&tgt)==0){
          FXASSERT(tgt);
          getApp()->event.type=SEL_DND_DROP;
          getApp()->event.window=getApp()->xdndTarget;
          tgt->handle(this,MKUINT(0,SEL_DND_DROP),&getApp()->event);
          getApp()->xdndFinishPending=FALSE;        // We just handled the callback
          dropped=TRUE;                             // We're sure, as its handled directly!
          }
 
        // Else send through protocol
        else{
          se.xclient.type=ClientMessage;
          se.xclient.display=getDisplay();
          se.xclient.message_type=getApp()->xdndDrop;
          se.xclient.format=32;
          se.xclient.window=getApp()->xdndTarget;
          se.xclient.data.l[0]=getApp()->xdndSource;
          se.xclient.data.l[1]=0;
          se.xclient.data.l[2]=getApp()->event.time;
          se.xclient.data.l[3]=0;
          se.xclient.data.l[4]=0;
          XSendEvent(getDisplay(),getApp()->xdndTarget,True,NoEventMask,&se);
          getApp()->xdndFinishPending=TRUE;         // Waiting for the drop confirmation!
          }
        
        // We wait till the drop has been processed...
        // The target should request the data at some point;
        // We should wait here till the target has processed the drop message
        // or some amount of timeout has expired....
        // Some error code should be returned if the timer expired...
        if(getApp()->xdndFinishPending){
          loops=1000;
          while(loops){
            if(XCheckIfEvent(getDisplay(),&se,matchevent,(char*)&getApp()->xdndFinished)){
              getApp()->dispatchEvent(se);
              if(se.xany.type==ClientMessage && se.xclient.data.l[0]==getApp()->xdndTarget){
//fprintf(stderr,"Got finish message!\n");
                getApp()->xdndFinishPending=FALSE;
                dropped=TRUE;
                break;
                }
 
              // We either got a selection request or a status message for a stale target.
              // Since there is thus still hope for a message, lets wait a bit longer...
              loops=1000;
              }
            fxsleep(20000);
            loops--;
            }
          }
        
        // We tried a drop
        nodrop=FALSE;
        }
      }
    
    // Didn't drop, or didn't get any response, so just send a leave
    if(nodrop){
      
      // If local, bypass protocol
      if(XFindContext(getDisplay(),getApp()->xdndTarget,getApp()->wcontext,(XPointer*)&tgt)==0){
        FXASSERT(tgt);
        getApp()->event.type=SEL_DND_LEAVE;
        getApp()->event.window=getApp()->xdndTarget;
        tgt->handle(this,MKUINT(0,SEL_DND_LEAVE),&getApp()->event);
        }
      
      // Else send through protocol
      else{
        se.xclient.type=ClientMessage;
        se.xclient.display=getDisplay();
        se.xclient.message_type=getApp()->xdndLeave;
        se.xclient.format=32;
        se.xclient.window=getApp()->xdndTarget;
        se.xclient.data.l[0]=getApp()->xdndSource;                  // Source window
        se.xclient.data.l[1]=0;
        se.xclient.data.l[2]=0;
        se.xclient.data.l[3]=0;
        se.xclient.data.l[4]=0;
        XSendEvent(getDisplay(),getApp()->xdndTarget,True,NoEventMask,&se);
        }
      }
      
    // Clean up
    if(getApp()->xdndTypeList) XFree(getApp()->xdndTypeList);
    getApp()->xdndTypeList=NULL;
    getApp()->xdndNumTypes=0;
    XDeleteProperty(getDisplay(),xid,getApp()->xdndTypes);
    getApp()->xdndSource=0;
    getApp()->xdndTarget=0;
    getApp()->xdndAccepts=FALSE;
    getApp()->xdndSendPosition=FALSE;
    getApp()->xdndStatusPending=FALSE;
    getApp()->xdndFinishPending=FALSE;
    getApp()->xdndStatusReceived=FALSE;
    getApp()->xdndWantUpdates=FALSE;
    getApp()->xdndVersion=XDND_PROTOCOL_VERSION;
    getApp()->xdndRect.x=0;
    getApp()->xdndRect.y=0;
    getApp()->xdndRect.w=0;
    getApp()->xdndRect.h=0;
    getApp()->xdndXPos=0;
    getApp()->xdndYPos=0;
    flags&=~FLAG_DRAGGING;
    }
#else
  FXASSERT(FALSE);
#endif
  return dropped;
  }


/*******************************************************************************/


// Delete window
FXWindow::~FXWindow(){
  if(prev) prev->next=next; else if(parent) parent->first=next;
  if(next) next->prev=prev; else if(parent) parent->last=prev; 
  if(parent && parent->focus==this) parent->focus=NULL;
  if(getApp()->refresher==this) getApp()->refresher=parent;
  if(getApp()->cursorWindow==this) getApp()->cursorWindow=parent;
  if(getApp()->focusWindow==this) getApp()->focusWindow=parent;
  if(getApp()->grabWindow==this) getApp()->grabWindow=NULL;
  if(parent) parent->recalc();
  delete accelTable;
  destroy(); 
  parent=(FXWindow*)-1;
  shell=(FXWindow*)-1;
  first=last=(FXWindow*)-1;
  next=prev=(FXWindow*)-1;
  focus=(FXWindow*)-1;
  defaultCursor=(FXCursor*)-1;
  dragCursor=(FXCursor*)-1;
  accelTable=(FXAccelTable*)-1;
  target=(FXObject*)-1;
  }

